# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Device Power Default Capability."""
import time

from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger

from gazoo_device.capabilities.interfaces import device_power_base

logger = gdm_logger.get_gdm_logger()

SUPPORTED_HUB_TYPES = ["cambrionix", "powerswitch", "unifi_switch", "ethernet_switch"]


class DevicePowerDefault(device_power_base.DevicePowerBase):
    """Base class for device_power."""

    def __init__(self,
                 device_name,
                 create_device_func,
                 hub_type,
                 props,
                 settable,
                 hub_name_prop,
                 port_prop,
                 wait_for_bootup_complete_fn,
                 switchboard_inst,
                 change_triggers_reboot=False):
        """Create an instance of the device_power capability.

        Args:
            device_name (str): name of the device this capability is attached to.
            create_device_func (func): create_device method.
            hub_type (str): type of switch for power cycling.
            props (dict): dictionary of device props from configuration file.
            settable (bool): whether or not the properties are settable.
            hub_name_prop (str):  name of the hub name property
            port_prop (str): name of the hub port property
            wait_for_bootup_complete_fn (func): a method that the capability can call to wait for
                                                a reboot to complete if triggered by a change.
            switchboard_inst (object): instance of switchboard capability
            change_triggers_reboot (bool): set change_triggers_reboot to TRUE if changing the
                                           power mode for the device causes a reboot.

        Raises:
            ValueError: if hub type not a known type.
        """
        super().__init__(device_name=device_name)

        self._hub = None
        self._create_device_func = create_device_func
        self._settable = settable
        self._hub_name_prop = hub_name_prop
        self._props = props
        self._port_prop = port_prop
        if hub_type not in SUPPORTED_HUB_TYPES:
            raise ValueError(
                "Hub Type {} not supported. Valid types: {}".format(
                    hub_type, SUPPORTED_HUB_TYPES))
        self._hub_type = hub_type

        # Set the properties
        self._dict_name = "optional" if self._settable else "persistent_identifiers"

        self._wait_for_bootup_complete_fn = wait_for_bootup_complete_fn
        self._switchboard = switchboard_inst
        self._change_triggers_reboot = change_triggers_reboot

    @decorators.CapabilityLogDecorator(logger, level=decorators.DEBUG)
    def health_check(self):
        """Checks that the capability is ready to use.

        Raises:
            CapabilityNotReadyError: if unable to create auxiliary device for power switching.
        """
        unset_props = []
        if self.hub_name is None:
            unset_props.append(self._hub_name_prop)
        if self.port_number is None:
            unset_props.append(self._port_prop)
        if unset_props:
            if self._settable:
                msg_format = ("If device is connected to {}, "
                              "set them via 'gdm set-prop {} <property> <value>'")
            else:
                msg_format = ("If device is connected to {}, "
                              "set them via 'gdm redetect {}")
            msg = msg_format.format(self._hub_type, self._device_name)
            error_msg = "properties {} are unset. ".format(" and ".join(unset_props)) + msg
            raise errors.CapabilityNotReadyError(msg=error_msg,
                                                 device_name=self._device_name)
        try:
            self._hub = self._create_device_func(self.hub_name)
        except errors.GazooDeviceError as err:
            raise errors.CapabilityNotReadyError(msg=str(err),
                                                 device_name=self._device_name)
        self._healthy = True

    @decorators.SettableProperty
    def hub_name(self):
        """Name of the hub the device is attached to."""
        return self._props[self._dict_name].get(self._hub_name_prop)

    @decorators.PersistentProperty
    def hub_type(self):
        """Device type of the hub."""
        return self._hub_type

    @decorators.DynamicProperty
    def port_mode(self):
        """port mode."""
        if not self.healthy:
            self.health_check()
        return self._hub.switch_power.get_mode(self.port_number)

    @decorators.SettableProperty
    def port_number(self):
        """Port number the device is attached to."""
        return self._props[self._dict_name].get(self._port_prop)

    @decorators.CapabilityLogDecorator(logger)
    def cycle(self, no_wait=False):
        """Power off then power on the device.

        Args:
            no_wait (bool):  Return before verifying boot up.
        """
        if not self.healthy:
            self.health_check()
        self.off()
        time.sleep(2)  # Small delay before calling power_on
        self.on(no_wait=no_wait)

    @decorators.CapabilityLogDecorator(logger)
    def off(self):
        """Turn off power to the device."""
        if not self.healthy:
            self.health_check()
        if self._change_triggers_reboot:
            self._switchboard.add_log_note(
                f"GDM triggered reboot via {self.hub_type} power change.")
            self._hub.switch_power.power_off(self.port_number)
            self._wait_for_bootup_complete_fn()
        else:
            self._switchboard.close_all_transports()
            self._hub.switch_power.power_off(self.port_number)

    @decorators.CapabilityLogDecorator(logger)
    def on(self, no_wait=False):
        """Turn on power to the device.

        Args:
            no_wait (bool):  Return before verifying boot up.
        """
        if not self.healthy:
            self.health_check()
        if self._change_triggers_reboot:
            self._switchboard.add_log_note(
                f"GDM triggered reboot via {self.hub_type} power change.")
        self._hub.switch_power.power_on(self.port_number)
        if not self._change_triggers_reboot:
            time.sleep(2)  # Small delay to give time for 'dev/tty' to populate
            self._switchboard.open_all_transports()

        if not no_wait:
            self._wait_for_bootup_complete_fn()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Fire Manager module.

Used for GDM CLI-specific commands and flags. Built to work with Python Fire.
"""
import codecs
import enum
import inspect
import json
import logging
import os.path
import pydoc
import re
import sys
import textwrap
import time
from gazoo_device import config
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import manager
from gazoo_device import gdm_logger
from gazoo_device import testbed
from gazoo_device.utility import parallel_utils
from gazoo_device.utility import usb_utils

logger = gdm_logger.get_gdm_logger()


class AttributeClassification(enum.Enum):
    """Types of class attributes recognized by GDM's "man" method."""
    CAPABILITY = "capability"
    DEPRECATED_METHOD = "deprecated method"
    DEPRECATED_PROPERTY = "deprecated property"
    HEALTH_CHECK = "health check"
    OTHER = "other"
    PRIVATE_METHOD = "private method"
    PROPERTY = "property"
    CLASS_PROPERTY = "class property"
    PUBLIC_METHOD = "public method"


HEALTHY_DEVICE_HEALTH = {
    "is_healthy": True,
    "unhealthy_reason": "",
    "err_type": "",
    "checks_passed": [],
    "properties": {}
}
MAX_TIME_TO_WAIT_FOR_INITATION = 5

_DOC_INDENT_SIZE = 4
# Capability attributes visible on device summary man page (such as "man cambrionix").
_VISIBLE_CAPABILITY_ATTRIBUTES = [AttributeClassification.PROPERTY,
                                  AttributeClassification.PUBLIC_METHOD]


class FireManager(manager.Manager):
    """Manages the setup and communication of Gazoo devices."""

    def __init__(self, debug=False, dev_debug=False, quiet=False):
        stream_debug = debug or dev_debug
        debug_level = logging.WARNING if quiet else logging.DEBUG

        super().__init__(debug_level=debug_level, stream_debug=stream_debug)

    def exec(self, identifier):
        """Alias for create_device with health checks disabled.

        Important: ensure the device is in a healthy state before using "exec" by running health
        checks via `gdm health-check device-1234`.
        Usage from the CLI: `gdm exec device-1234 - shell "echo 'foo'"`.

        Args:
            identifier (str): The identifier string (name) which specifies the device.

        Returns:
            Object: The device found and created by the identifier specified.
        """
        return self.create_device(identifier, make_device_ready="off")

    def get_persistent_prop_devices(self, devices):
        """Gets the persistent properties of one or more devices and returns a json response.

        Args:
          devices (list): Device identifiers to get properties of.

        Returns:
          str: json formatted persistent properties, e.g.
            {
              'device-1234': {
                'build_date': '180124',
                'ftdi_serial_number': '1f824023'
              },
              'device-5678': {},
              ...
            }

        Note:
          This assumes the provided devices are healthy. If properties are unable to be fetched,
          an empty json object will be returned for the device for which the properties could not
          be retrieved.
        """
        logger.setLevel(logging.ERROR)  # silence logging to reduce CLI output
        devices_props = {}
        if isinstance(devices, str):
            devices = devices.split(",")

        for device_name in devices:
            persistent_props = {}
            try:
                device_props = self.get_device_configuration(device_name)
                persistent_props = device_props.get("persistent", {})
            except errors.GazooDeviceError:
                pass  # unhealthy devices will have empty props
            devices_props[device_name] = persistent_props

        return json.dumps(devices_props)

    def get_prop(self, device_name, prop=None):
        """Prints the device properties.

        Args:
          device_name (str): identifier for device.
          prop (str): identifier for property.
        """
        format_line = "  {:22s}  {}"

        if prop is not None:  # If requested a single property...
            value = self.get_device_prop(device_name, prop)
            logger.info(format_line.format(prop, str(value)))
        else:
            props_dicts = self.get_device_prop(device_name, prop)
            property_types = ["persistent", "settable", "dynamic"]
            for property_type in property_types:
                title = "{} Properties:".format(property_type.capitalize())
                logger.info("")
                logger.info(title)
                for prop_name in sorted(props_dicts[property_type]):
                    prop_value = props_dicts[property_type][prop_name]
                    if isinstance(prop_value, list) and len(prop_value) > 1:
                        logger.info(format_line.format(prop_name, ""))
                        for value in prop_value:
                            if callable(value):
                                value = value.__name__
                            logger.info(
                                "\t\t\t  {}".format(str(value)))
                        logger.info("")
                    elif isinstance(prop_value, dict) and len(prop_value) > 1:
                        logger.info(format_line.format(prop_name, ""))
                        for key in sorted(prop_value):
                            logger.info("\t\t\t  {:25s} {!r}".format(key, prop_value[key]))
                        logger.info("")

                    else:
                        logger.info(format_line.format(prop_name, prop_value))

    def health_check(self, identifier, recover=False):
        """CLI command for running device health checks.

        Usage from the CLI: `gdm health-check device-1234`.

        Args:
            identifier (str): The identifier string (name) which specifies the device.
            recover (bool): whether to automatically recover from health check failures.
        """
        make_device_ready_setting = "check_only"
        if recover:
            make_device_ready_setting = "on"
        device = self.create_device(identifier, make_device_ready="off")
        try:
            device.make_device_ready(setting=make_device_ready_setting)
        finally:
            device.close()

    @classmethod
    def helpfull(cls):
        """Prints a general overview of GDM's features. Invoked by 'man' without arguments."""
        description = textwrap.dedent("""
        Gazoo Device Manager (gdm) command line interface is a single utility for control of a
        variety of smart devices.

        Commands:
            The CLI is dynamic, meaning commands are generated from API methods (see
            https://github.com/google/python-fire).
            There are several groups of commands.
                "Manager" commands operate on local config files on the pc.
                These commands are generated from API methods in the Manager class.
                    The command may talk to devices (detect), but they don't change the state of
                    the device (i.e. read-only)
                    The command may change the config files stored on the pc, to detect new
                    devices, add an alias, etc.
                    To see a list of available manager commands, run "gdm".
                "Device" commands talk to or modify a Gazoo device.
                These commands are generated from API methods in device classes.
                    This includes upgrade, reboot, etc.
                    In general you enter a device command as "gdm issue <device_name> -
                    <command>"
                    Examples::
                    gdm issue cambrionix-zdwm - reboot
                    gdm issue device-1234 - upgrade --build-number 1.0d1038
                    To see the list of device commands available for a device, run "gdm man
                    <device_type>"
                    For example:: gdm man cambrionix

            You can get more details on a particular command at the command line with:
            "gdm <command> -h"
            For example::  gdm detect -h

            You can pass in flags to the CLI with:
            "gdm --<flag> - <command>"
            For example:: gdm --debug - devices
            To see a list of available flags, run "gdm -h"

        Supported device types:
        Primary:
        {}

        Auxiliary:
        {}

        Virtual:
        {}

        To explore available device functionality through the dynamic CLI, you will need a device
        attached to your host.

        Use "gdm man" to access static documentation (no devices necessary, but has limitations):
            "gdm man <device_type>" to see all functionality supported by a device.
            "gdm man <device_type> --deprecated" to see just deprecated functionality for a device.
            "gdm man <device_type> <class_attribute>" for detailed documentation.
        """).format(cls._indent_doc_lines(cls.get_supported_primary_device_types()),
                    cls._indent_doc_lines(cls.get_supported_auxiliary_device_types()),
                    cls._indent_doc_lines(cls.get_supported_virtual_device_types()))[1:-1]
        logger.info(description)

    def issue(self, identifier, **kwargs):
        """Alias for create_device with health checks enabled by default.

        Usage from the CLI: `gdm issue device-1234 - shell "echo 'foo'"`.

        Args:
            identifier (str): The identifier string (name) which specifies the device.
            **kwargs (dict): keyword-only arguments passed on to create_device:
                "log_file_name" (str) -- a string log file name to use for log results.
                "log_directory" (str) -- a directory path to use for storing log file.
                "make_device_ready" (str) -- health check setting ("on", "check_only", "off").

        Returns:
            Object: The device found and created by the identifier specified.
        """
        return self.create_device(identifier,
                                  log_file_name=kwargs.get("log_file_name"),
                                  log_directory=kwargs.get("log_directory"),
                                  make_device_ready=kwargs.get("make_device_ready", "on"))

    def log(self, device_name, log_file_name=None, duration=2000):
        """Streams device logs to stdout.

        Args:
          device_name (str): device identifier.
          log_file_name (str): log_file_name. Used for testing purposes.
          duration (float): how long to stream logs for.

        Raises:
          GazooDeviceError: if unable to initiate log file to stream in 10 seconds.
        """
        logger.info("Streaming logs for max {}s".format(duration))
        device = self.create_device(device_name, log_file_name=log_file_name)
        # Disable log rotation feature
        if hasattr(type(device), "switchboard"):
            device.switchboard.set_max_log_size(0)

        try:
            # Wait up to 10 seconds for log file to be created
            end_time = time.time() + MAX_TIME_TO_WAIT_FOR_INITATION
            while time.time() < end_time:
                if os.path.exists(device.log_file_name):
                    break
                time.sleep(0.001)
            else:
                raise errors.GazooDeviceError(
                    "Streaming logs for {} failed. "
                    "Log file not created within {} seconds".format(
                        device_name, MAX_TIME_TO_WAIT_FOR_INITATION))

            start_time = time.time()
            end_time = start_time + duration
            # Open log file and process log file
            with codecs.open(device.log_file_name,
                             "r",
                             encoding="utf-8",
                             errors="replace") as log_file:

                while time.time() < end_time:
                    line = log_file.readline()
                    if line:
                        sys.stdout.write(line)
                        sys.stdout.flush()
                    else:
                        time.sleep(0.001)

        finally:
            sys.stdout.flush()
            device.close()

    def make_devices_ready(self, devices, testing_props=None, aggressive=False):
        """Makes one or more devices ready and returns a json response.

        Args:
          devices (list): Devices identifiers to make ready.
          testing_props (dict): Properties of the testbed used for testing.
          aggressive (bool): Re-flash the device with a valid build if recovery fails.

        Returns:
          str: json formatted device health, e.g.
            {
              'device-1234': {
                'is_healthy': true or false if the device is healthy or not,
                'unhealthy_reason': error message if device is unhealthy,
                'err_type': type of exception raised, if any
              },
              'device-5678': {
                'is_healthy': ...,
                'unhealthy_reason': ...,
                'err_type': ...
              },
              ...
            }
        """
        logger.setLevel(logging.ERROR)  # silence logging to reduce CLI output
        combined_results = {}
        if isinstance(devices, str):
            devices = devices.split(",")

        # create device instances and construct parameter dicts
        device_instances = []
        parameter_dicts = {}
        for device_name in devices:
            try:
                device = self.create_device(device_name, make_device_ready="off")
            except errors.GazooDeviceError as err:
                combined_results[device_name] = self._construct_health_dict_from_exception(err)
            else:
                device_instances.append(device)

                # pass make_device_ready setting to support aggressive recovery
                setting = "flash_build" if aggressive else "on"
                parameter_dicts[device.DEVICE_TYPE] = {"setting": setting}

        # execute testbed health checks if testing props provided
        # property keys that exist in Testbed.PROP_TO_HEALTH_CHECK trigger health checks
        if testing_props:
            try:
                testbed.Testbed(device_instances, testing_props).make_testbed_ready()
            except errors.GazooDeviceError as err:
                combined_results["testbed"] = self._construct_health_dict_from_exception(err)
            else:
                combined_results["testbed"] = HEALTHY_DEVICE_HEALTH

        # execute manager method with each device instance in parallel
        if device_instances:
            results = parallel_utils.parallel_process(
                action_name="make_device_ready",
                fcn=self._make_devices_ready_single_device,
                devices=device_instances,
                parameter_dicts=parameter_dicts)

            # combine results of parallel calls
            for result in results:
                if isinstance(result, dict):
                    combined_results.update(result)
                else:
                    logger.info(result)

        # device instances no longer needed
        for device in device_instances:
            device.close()

        return json.dumps(combined_results)

    @classmethod
    def man(cls, device_type=None, class_attr_name=None, deprecated=False):
        """Prints documentation without reliance on device instance creation.

        Args:
            device_type (str): device type.
            class_attr_name (str): name of the class attribute to display documentation for.
            deprecated (bool): display only deprecated methods and properties.

        Raises:
            GazooDeviceError: if called for documentation on a nested class attribute,
                             such as "touch.zoom".

        Note: dynamic device documentation ("gdm <command> --help") is more complete, but
              requires a device attached to the host to generate documentation.
        """
        if device_type and "." in device_type and class_attr_name is None:  # "man device.method"
            device_type, class_attr_name = device_type.split(".", 1)
        if class_attr_name and "." in class_attr_name:
            raise errors.GazooDeviceError(
                "Static documentation for nested class attributes is not available. "
                "Refer to dynamic CLI documentation instead.")

        if device_type is None and class_attr_name is None:
            cls.helpfull()
        elif class_attr_name is None:
            cls._man_device(device_type, deprecated)
        else:
            cls._man_class_attribute(device_type, class_attr_name)

    def print_usb_info(self):
        """Prints the usb_info dictionary in a human readable form."""
        values = usb_utils.get_address_to_usb_info_dict().values()
        logger.info("{} USB connections found.".format(len(values)))
        for num, usb_info_dict in enumerate(values):
            keys = usb_info_dict.get_properties()
            logger.info("Connection {}:".format(num))
            for key in keys:
                logger.info("\t{:15} {:15}".format(key, str(getattr(usb_info_dict, key))))

    @classmethod
    def _classify_attribute(cls, class_attr):
        """Classifies the class attribute."""
        class_attr = decorators.unwrap(class_attr)
        if isinstance(class_attr, decorators.CapabilityProperty):
            return AttributeClassification.CAPABILITY
        if (hasattr(class_attr, "__deprecated__")
                or hasattr(getattr(class_attr, "fget", None), "__deprecated__")):
            if inspect.isroutine(class_attr):
                return AttributeClassification.DEPRECATED_METHOD
            if isinstance(class_attr, property):
                return AttributeClassification.DEPRECATED_PROPERTY
        if isinstance(class_attr, config.CLASS_PROPERTY_TYPES):
            return AttributeClassification.CLASS_PROPERTY
        if isinstance(class_attr, property):
            return AttributeClassification.PROPERTY
        if inspect.isroutine(class_attr):
            if hasattr(class_attr, "__health_check__"):
                return AttributeClassification.HEALTH_CHECK
            if class_attr.__name__.startswith('_'):
                return AttributeClassification.PRIVATE_METHOD
            return AttributeClassification.PUBLIC_METHOD
        return AttributeClassification.OTHER

    def _construct_health_dict_from_exception(self, exc):
        """Constructs a dictionary containing information about an unhealty device's issues.

        Args:
          exc (Exception): the exception raised that is causing the device's issues

        Returns:
          str: json formatted device health, e.g.
            {
              device-1234: {
                'is_healthy': true,
                'unhealthy_reason': "",
                'err_type': "",
                'checks_passed': [],
                'properties': {}
              }
            }
        """
        device_health = {}
        device_health["is_healthy"] = False
        device_health["unhealthy_reason"] = str(exc)
        device_health["err_type"] = type(exc).__name__
        device_health["checks_passed"] = getattr(exc, "checks_passed", [])
        device_health["properties"] = getattr(exc, "properties", {})
        return device_health

    @classmethod
    def _indent_doc_lines(cls, doc_lines, indent=_DOC_INDENT_SIZE):
        """Indents docstring lines."""
        indent_str = " " * indent
        return "\n".join(indent_str + line for line in doc_lines)

    def _make_devices_ready_single_device(self, device, parameter_dict):
        """Execute make_device_ready for a single device.

        Args:
          device (GazooDeviceBase or AuxiliaryDeviceBase): device to execute make_device_ready on
          parameter_dict (dict): dictionary storing the setting to pass to make_device_ready.

        Returns:
          dict: device health, e.g.
            'device-1234': {
              'is_healthy': true or false if the device is healthy or not,
              'unhealthy_reason': error message if device is unhealthy,
              'err_type': type of exception raised, if any
            }

        Note:
          Intended to be executed in parallel by parallel_utils, triggered by the
          make_devices_ready method.
        """
        device_health = HEALTHY_DEVICE_HEALTH
        try:
            device.make_device_ready(setting=parameter_dict.get("setting"))
        except errors.GazooDeviceError as err:
            device_health = self._construct_health_dict_from_exception(err)

        device_health["logs"] = device.log_file_name
        return {device.name: device_health}

    @classmethod
    def _man_class_attribute(cls, device_type, class_attr_name):
        """Prints class attribute documentation. Invoked by 'man' with 2 args (device, attribute).

        Args:
            device_type (str): device type.
            class_attr_name (str): name of the class attribute to display documentation for.

        Raises:
            GazooDeviceError: if the requested class attribute isn't present for the device type.
        """
        # Find the device class, fetch the requested attribute, and determine what the attribute is
        device_class = cls.get_supported_device_class(device_type)
        if not hasattr(device_class, class_attr_name):
            raise errors.GazooDeviceError("{} ({}) does not have attribute {!r}".format(
                device_type, device_class, class_attr_name))
        attribute = getattr(device_class, class_attr_name)
        classification = cls._classify_attribute(attribute)
        description = classification.value

        if classification == AttributeClassification.CAPABILITY:
            capability_classes = list(attribute.capability_classes)
            attribute = capability_classes[0]
            description += "; flavor: {}".format(attribute.__name__)

            if len(capability_classes) > 1:
                # Capabilities can have multiple flavors in one device class.
                # The flavor used is determined based on device firmware at runtime.
                # Print pydoc just for the first class to avoid clutter and add a warning.
                flavors = [a_cls.__name__ for a_cls in capability_classes]
                logger.warning(
                    "{flavor_num} flavors ({flavors}) of capability `{cap_name}` are available "
                    "for {device_type}.\nShowing documentation for flavor {first_flavor}.\n"
                    .format(flavor_num=len(capability_classes),
                            flavors=flavors,
                            cap_name=class_attr_name,
                            device_type=device_type,
                            first_flavor=attribute.__name__))

        doc_title = "Manual for {class_name}.{attribute_name} ({description})\n".format(
            class_name=device_class.__name__,
            attribute_name=class_attr_name,
            description=description)
        pydoc_lines = pydoc.render_doc(attribute).splitlines()
        doc = doc_title + "\n".join(pydoc_lines[1:])  # Replace pydoc's title
        logger.info(doc)

    @classmethod
    def _man_device(cls, device_type, deprecated):
        """Prints supported device features. Invoked by 'man' with 1 argument (device type)."""
        device_class = cls.get_supported_device_class(device_type)

        # Group device class attributes into properties, health checks, methods, and capabilities
        capability_name_to_classes = {}
        property_names = []
        class_properties = []
        public_methods_names = []
        health_check_names = []
        deprecated_methods = []
        deprecated_properties = []
        for name, attribute in inspect.getmembers(device_class):
            classification = cls._classify_attribute(attribute)
            if classification == AttributeClassification.PROPERTY:
                property_names.append(name)
            elif classification == AttributeClassification.CLASS_PROPERTY:
                if not name.startswith("_"):
                    class_properties.append(f"{name}: {attribute}")
            elif classification == AttributeClassification.HEALTH_CHECK:
                health_check_names.append(name)
            elif classification == AttributeClassification.PUBLIC_METHOD:
                public_methods_names.append(name)
            elif classification == AttributeClassification.CAPABILITY:
                capability_name_to_classes[name] = list(attribute.capability_classes)
            elif classification == AttributeClassification.DEPRECATED_METHOD:
                deprecated_methods.append(f"{name} ({attribute.__deprecated__})")
            elif classification == AttributeClassification.DEPRECATED_PROPERTY:

                # Parse alias from deprecated property docstring
                match = re.search(r'See "(?P<alias>.*)".', attribute.__doc__)
                alias = match.group("alias")
                deprecated_properties.append(f"{name} ({alias})")

        # Generate a summary of supported capability methods and properties for each capability
        capability_lines = []
        for cap_name, cap_classes in capability_name_to_classes.items():
            # There may be several flavors of a capability for a device class.
            for cap_class in cap_classes:
                capability_lines.append("{} ({})".format(cap_name, cap_class.__name__))
                methods_and_props = [
                    name for name, member in inspect.getmembers(cap_class)
                    if cls._classify_attribute(member) in _VISIBLE_CAPABILITY_ATTRIBUTES
                ]
                indented_lines = [" " * _DOC_INDENT_SIZE + line for line in methods_and_props]
                capability_lines.extend(indented_lines)

        docs_capabilities = cls._indent_doc_lines(capability_lines)
        docs_methods = cls._indent_doc_lines(public_methods_names)
        docs_health_checks = cls._indent_doc_lines(health_check_names)
        docs_properties = cls._indent_doc_lines(property_names)
        docs_class_properties = cls._indent_doc_lines(class_properties)
        docs_deprecated_methods = cls._indent_doc_lines(deprecated_methods)
        docs_deprecated_properties = cls._indent_doc_lines(deprecated_properties)

        if deprecated:
            template = textwrap.dedent("""
            Deprecated manual for device type '{device_type}' (class {device_class})

            Deprecated methods:
            {deprecated_methods}

            Deprecated properties:
            {deprecated_properties}

            Use 'gdm man {device_type} <deprecated_attribute>' for detailed documentation.""")[1:]
        else:
            template = textwrap.dedent("""
            Manual for device type '{device_type}' (class {device_class})

            {ownership}
            Supported capabilities:
            {capabilities}

            Supported methods:
            {methods}

            Supported health check methods:
            {health_checks}

            Supported properties:
            {properties}

            Class properties:
            {class_properties}

            Use 'gdm man {device_type} <class_attribute>' for detailed documentation.""")[1:]

        ownership = ""
        owner = getattr(device_class, "_OWNER_LDAP", None)
        if owner:
            ownership = "Owned by {}\n".format(owner)

        doc = template.format(device_type=device_type,
                              device_class=device_class,
                              capabilities=docs_capabilities,
                              methods=docs_methods,
                              ownership=ownership,
                              health_checks=docs_health_checks,
                              properties=docs_properties,
                              class_properties=docs_class_properties,
                              deprecated_methods=docs_deprecated_methods,
                              deprecated_properties=docs_deprecated_properties)
        logger.info(doc)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Enables logging from multiple processes.

Step 1: Initialize a queue in the main process.
Step 2: Pass that queue to a LoggingThread object.
Step 3: Add handlers (StreamHandler, FileHandler, etc) to the LoggingThread.
Step 4: In each process, initialize a QueueHandler and add it to a Logger.
Step 5: Send messages to the Logger.

Messages travel from the Logger, to the QueueHandler, into the LoggingThread's
child thread via the queue, then into the destination handlers.

Adapted from the implementation in Python 3.5.
"""

from __future__ import absolute_import
from __future__ import print_function
import gc
import logging
import sys
import threading

SYNC_TIMEOUT = 0.25
TERMINATE_TIMEOUT = 2


class _Sentinel(object):
    """Sentinels to put in the queue to signal certain events."""
    TERMINATE = 0
    SYNC = 1


class DisablePeriodicGC(object):
    """Context manager for disabling & re-enabling periodic garbage collection.

    Note: will not disable periodic GC if called during periodic garbage collection.
    See b/143321081 for why logging during __del__ can deadlock if periodic GC is not disabled.
    """
    is_periodic_gc_active = False

    @classmethod
    def _track_gc_state(cls, phase, _):
        if phase == "start":
            cls.is_periodic_gc_active = True
        elif phase == "stop":
            cls.is_periodic_gc_active = False

    def __enter__(self):
        """Enter the context manager & disable periodic GC."""
        self._was_periodic_gc_enabled = gc.isenabled()
        if self._was_periodic_gc_enabled and not self.is_periodic_gc_active:
            gc.disable()

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Exit the context manager & re-enable periodic GC (if it had been enabled before)."""
        if self._was_periodic_gc_enabled and not self.is_periodic_gc_active:
            gc.enable()


if hasattr(gc, "callbacks"):  # Available in Python 3 only
    gc.callbacks.append(DisablePeriodicGC._track_gc_state)


class QueueHandler(logging.Handler):
    """Receives all messages destined for the handlers in LoggingThread."""

    def __init__(self, queue):
        logging.Handler.__init__(self)
        self._queue = queue

    def emit(self, record):
        try:
            self.format(record)
            record.msg = record.message
            record.args = None
            record.exc_info = None

            with DisablePeriodicGC():
                self._queue.put_nowait(record)

        except Exception:
            self.handleError(record)


class LoggingThread(object):
    """Runs in main process and pulls log messages from the shared queue."""

    def __init__(self, queue):
        self._handlers = []
        self._queue = queue
        self._thread = None
        self._synchonize_event = threading.Event()

    def add_handler(self, handler):
        """Adds a logging handler to the LoggingThread.

        Args:
            handler (logging.Handler): logging handler to add to the logging queue handlers.

        The handler will receive messages that are placed on the shared queue by QueueHandlers.
        """

        self._handlers.append(handler)

    def remove_handler(self, handler):
        """Removes the given logging handler from the LoggingThread."""

        try:
            self._handlers.remove(handler)
        except ValueError:
            pass

    def start(self):
        """Starts the child thread, which pulls messages from the queue."""

        self._thread = threading.Thread(target=self._run,
                                        args=[self._queue, self._synchonize_event])
        self._thread.daemon = True
        self._thread.start()

    def stop(self):
        """Signals to the child thread to stop by putting None on the queue.

        Raises:
          RuntimeError: Thread alive after attempt to stop it.
        """

        if self._thread:
            with DisablePeriodicGC():
                self._queue.put_nowait(_Sentinel.TERMINATE)
            self._thread.join(timeout=TERMINATE_TIMEOUT)

            if self._thread.is_alive():
                raise RuntimeError('Failed to stop LoggingThread in {}s'.format(TERMINATE_TIMEOUT))

        self._thread = None

    def sync(self, timeout=SYNC_TIMEOUT):
        """Put a sentinel in the message queue and block until the logging thread reaches it.

        Args:
            timeout (float): maximum time to wait for logging thread to reach the synchronization
                             sentinel.

        Note:
            Does nothing in child (forked) processes because logging thread is not copied over
            by os.fork().
            Also does nothing if called in any thread other than the main thread.
        """
        if (self._thread is not None and self._thread.is_alive()
                and isinstance(threading.current_thread(), threading._MainThread)):
            with DisablePeriodicGC():
                self._queue.put_nowait(_Sentinel.SYNC)
            if not self._synchonize_event.wait(timeout):
                print("Warning: Logging thread did not reach the synchronization sentinel in {}s"
                      .format(timeout), file=sys.stderr)
            self._synchonize_event.clear()

    def _run(self, queue, synchronize_event):
        """Runs as a child thread.

        Args:
            queue (multiprocessing.Queue): multiprocess queue to receive log messages from.
            synchronize_event (threading.Event): event used for synchronization with the main
                                                 execution thread.

        Continuously receiving messages placed on the
        queue by QueueHandlers. Passes messages to each of the LoggingThread's
        handlers. Stops when None is found on the queue.
        """

        while True:
            record = queue.get()

            if record == _Sentinel.TERMINATE:
                break
            elif record == _Sentinel.SYNC:
                synchronize_event.set()
            else:
                for handler in self._handlers:
                    if record.levelno >= handler.level:
                        handler.handle(record)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""adb flavor of file transfer capability."""
from __future__ import absolute_import
import os

from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import file_transfer_base
from gazoo_device.utility import adb_utils

logger = gdm_logger.get_gdm_logger()


class FileTransferAdb(file_transfer_base.FileTransferBase):
    """adb file transfer capability."""

    def __init__(self, device_name, add_log_note_fn, communication_address):
        """Initialize adb file transfer capability.

        Args:
            device_name (str): name of the device using the capability.
            add_log_note_fn (func): add output to log file.
            communication_address (str): device serial number
        """
        super(FileTransferAdb, self).__init__(device_name=device_name)
        self._add_log_note_fn = add_log_note_fn
        self._communication_address = communication_address

    @decorators.CapabilityLogDecorator(logger)
    def recv_file_from_device(self, src, dest="./"):
        """Copies srcs on device to dest on host.

        Args:
            src (str): file path on device to copy to host.
            dest (str): destination path on host computer.

        Raises:
            GazooDeviceError: if destination directory doesn't exist or copy failed.

        Note:
            If no dest is provided, the file will be copied to the current working directory
                on the host computer.
        """
        destination_dir = os.path.dirname(dest)
        if destination_dir != "." and not os.path.exists(destination_dir):
            raise errors.GazooDeviceError("Device {} receive from device failed. "
                                          "Destination directory {} doesn't appear to exist.".
                                          format(self._device_name,
                                                 destination_dir))

        logger.info("{} receiving from device. Source: {} Destination: {}",
                    self._device_name, src, dest)
        output = adb_utils.pull_from_device(self._communication_address,
                                            src,
                                            destination_path=dest)
        for line in output.splitlines():
            self._add_log_note_fn(line + "\n")
        if not os.path.exists(dest):
            raise errors.GazooDeviceError(
                "Final file {} does not exist".format(dest))

    @decorators.CapabilityLogDecorator(logger)
    def send_file_to_device(self, src, dest):
        """Copies src from host to dest on the device.

        Args:
            src (str): local file path on host computer.
            dest (str): file path on device where the file should be copied to.

        Raises:
            GazooDeviceError: if source file doesn't exist or copy failed.
        """
        if not os.path.exists(src):
            raise errors.GazooDeviceError("Device {} send to device failed. "
                                          "Source file {} doesn't appear to exist.".
                                          format(self._device_name, src))

        logger.info("{} sending file(s) to device. Source: {} Destination: {}",
                    self._device_name, src, dest)
        try:
            output = adb_utils.push_to_device(self._communication_address, src, dest)
            for line in output.splitlines():
                self._add_log_note_fn(line + "\n")
        except RuntimeError as err:
            raise errors.GazooDeviceError("Unable to copy {} to {} on device. "
                                          "Error: {!r}".
                                          format(src,
                                                 dest,
                                                 err))



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of the switch_power_usb_default capability."""
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import switch_power_base

logger = gdm_logger.get_gdm_logger()

SYNC = "sync"
OFF = "off"


class SwitchPowerUsbDefault(switch_power_base.SwitchPowerBase):
    """Definition of the switch_power_usb_default capability."""
    _REQUIRED_COMMANDS = ["POWER_ON", "POWER_OFF", "GET_MODE"]
    _REQUIRED_REGEXS = ["GET_MODE_REGEX"]

    def __init__(self,
                 shell_fn,
                 regex_shell_fn,
                 command_dict,
                 regex_dict,
                 device_name,
                 serial_number,
                 total_ports):
        """Create an instance of the default hub power capability.

        Args:
            shell_fn (func): Function for calling shell on the device.
            regex_shell_fn (func): Function for calling shell_with_regex on the device.
            command_dict (dict): A dictionary containing the command used for each method.
            regex_dict (dict): A dictionary containing the response regex to use to
                               determine the return value for the property.
            device_name (str): name of the device this capability is attached to.
            serial_number (str): serial number of device this capability is attached to.
            total_ports (int): number of ports on device
        """
        super().__init__(device_name=device_name)
        self._shell_fn = shell_fn
        self._regex_shell_fn = regex_shell_fn
        self._command_dict = command_dict
        self._regex_dict = regex_dict
        self._serial_number = serial_number
        self._total_ports = total_ports

        self.validate_required_keys(self._REQUIRED_COMMANDS, command_dict, "command_dict")
        self.validate_required_keys(self._REQUIRED_REGEXS, regex_dict, "regex_dict")

    @decorators.PersistentProperty
    def supported_modes(self):
        """Get the USB power modes supported by the USB hub."""
        return [OFF, SYNC]

    def get_all_ports_mode(self):
        """Gets the USB mode for all ports on this hub.

        Returns:
          list: Returns a list of port modes with port number as index.
        """
        mode_list = ['']
        for port in range(1, self._total_ports + 1):
            mode_list.append(self.get_mode(port=port))
        return mode_list

    def get_mode(self, port):
        """Get the USB mode for specified port.

        Args:
          port (int): The port to get mode.

        Returns:
            str: USB port mode settings 'off', 'sync'

        Raises:
          GazooDeviceError: invalid port.
        """
        port = int(port)
        self._validate_port("get_mode", port)
        cmd = self._command_dict["GET_MODE"].format(self._serial_number, port)
        result = self._regex_shell_fn(command=cmd,
                                      regex=self._regex_dict["GET_MODE_REGEX"],
                                      command_name="get_mode",
                                      tries=5)
        if "ON" in result.upper():
            return SYNC
        return OFF

    @decorators.CapabilityLogDecorator(logger)
    def power_on(self, port, data_sync=True):
        """This command powers on the port specified.

        Args:
            port (int): identifying which hub port to power on
            data_sync (bool): Ignored in default implementation

        Raises:
           GazooDeviceError: if data_sync is False or if a valid port is not provided.
        """
        port = int(port)
        if not data_sync:
            raise errors.GazooDeviceError(
                "Device {} {} power_on failed. "
                "data_sync=False not supported in default implementation.".
                format(self._device_name, self._serial_number))

        self._validate_port("power_on", port)
        logger.debug("{} Powering on usb port {}".format(self._device_name, port))
        self._shell_fn(self._command_dict["POWER_ON"].format(self._serial_number, port))

    @decorators.CapabilityLogDecorator(logger)
    def power_off(self, port):
        """This command powers off the port specified.

        Args:
            port (int): identifies which hub port to power off

        Raises:
           GazooDeviceError: if operation failed or if a valid port is not provided.
        """
        port = int(port)
        self._validate_port("power_off", port)
        logger.debug("{} Powering off usb port {}".format(self._device_name, port))
        self._shell_fn(self._command_dict["POWER_OFF"].format(self._serial_number, port))

    @decorators.CapabilityLogDecorator(logger)
    def set_all_ports_mode(self, mode):
        """Sets all USB hub ports to the mode specified.

        Args:
          mode (str): USB hub mode to set. The mode must be in one of the supported_modes
                      Example, 'off', 'sync'

        Raises:
          GazooDeviceError: invalid mode.
        """
        self._validate_mode(mode)
        for port in range(1, self._total_ports + 1):
            self.set_mode(mode=mode, port=port)

    @decorators.CapabilityLogDecorator(logger)
    def set_mode(self, mode, port):
        """Sets the specified USB port to the mode specified.

        Args:
          mode (str): USB mode to set. Options: 'off', 'sync'
          port (int): The port to set.

        Raises:
          GazooDeviceError: invalid mode or port.
        """
        port = int(port)
        mode = mode.lower()
        self._validate_port("set_mode", port)
        self._validate_mode(mode)

        if mode in OFF:
            self.power_off(port)
        else:  # "sync"
            self.power_on(port)

    def _validate_mode(self, mode):
        """Verify mode given resides in the valid mode list.

        Args:
            mode (str): mode e.g. 'off', 'sync'

        Raises:
            GazooDeviceError: if mode given is not found in the valid mode list.
        """
        if mode not in list(self.supported_modes):
            raise errors.GazooDeviceError("Mode {} not found in supported modes {!r}".
                                          format(mode,
                                                 self.supported_modes))

    def _validate_port(self, method_name, port):
        """Verify port given is within the bounds defined.

        Args:
            method_name (str): name of method requesting validation.
            port (int): value provided must be >= 1 and <= _total_ports.

        Raises:
            GazooDeviceError: if port given is not within the bounds defined.
        """
        if not port or port > self._total_ports:
            raise errors.GazooDeviceError("Device {} {} failed. "
                                          "Port {} is invalid.".
                                          format(self._device_name, method_name, port))



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Docker flavor of file transfer capability (for devices using PTY transports)."""
from __future__ import absolute_import
import os

from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import file_transfer_base
from gazoo_device.utility import host_utils

logger = gdm_logger.get_gdm_logger()


class FileTransferDocker(file_transfer_base.FileTransferBase):
    """Docker-based file transfer capability."""

    def __init__(self, docker_container, device_name):
        """Initialize docker file transfer capability.

        Args:
            docker_container (str): name or id of the device docker container using the capability.
            device_name (str): name of the device using the capability
        """

        super().__init__(device_name=device_name)
        self._docker_container = docker_container

    @decorators.CapabilityLogDecorator(logger)
    def send_file_to_device(self, src, dest):
        """Copy a file from the local host to the device.

        Args:
            src (str): Absolute path of the source file on the host.
            dest (str): Absolute path of the destination on the device

        Raises:
            GazooDeviceError: if the file transfer fails for any reason,
               or if the specified mode is invalid.

        Note:
           dest may be either a directory or a file name.
        """
        if not os.path.exists(src):
            raise errors.GazooDeviceError("Device {} send to device failed. "
                                          "Source file {} doesn't appear to exist.".
                                          format(self._device_name, src))

        logger.info("{} sending file to device. Source: {} Destination: {}",
                    self._device_name, src, dest)
        host_utils.docker_cp_to_device(docker_container=self._docker_container,
                                       local_file_path=src,
                                       container_file_path=dest)

    @decorators.CapabilityLogDecorator(logger)
    def recv_file_from_device(self, src, dest="./"):
        """Copy a file from the device to the local host.

        Args:
            src (str): Absolute path of the source file on the device.
            dest (str): Absolute path of the destination on the host

        Raises:
          GazooDeviceError: if the file transfer fails for any reason.
          RuntimeError: if the final file does not exist

        Note:
         dest can be directory or a file name.
        """
        dest_dir = os.path.dirname(dest)
        if dest_dir != "." and not os.path.exists(dest_dir):
            raise errors.GazooDeviceError("Device {} receive from device failed. "
                                          "Destination directory {} doesn't appear to exist.".
                                          format(self._device_name,
                                                 dest_dir))

        logger.info("{} receiving file from device. Source: {} Destination: {}",
                    self._device_name, src, dest)
        host_utils.docker_cp_from_device(docker_container=self._docker_container,
                                         local_file_path=dest,
                                         container_file_path=src)
        if not os.path.exists(dest):
            raise RuntimeError(
                "Final file {} does not exist".format(dest))



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of the switch_power_dli_powerswitch capability."""
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import switch_power_base

logger = gdm_logger.get_gdm_logger()


OFF = 'off'
ON = 'on'


class SwitchPowerDliPowerswitch(switch_power_base.SwitchPowerBase):
    """Definition of the switch_power_dli_powerswitch capability."""
    _REQUIRED_COMMANDS = ["ADJUST_PORTS_MODE"]
    _REQUIRED_HEADERS = ["GET_PROP", "SET_PROP"]

    def __init__(self,
                 http_fn,
                 ip_address,
                 command_dict,
                 headers_dict,
                 device_name,
                 total_ports):
        """Create an instance of the powerswitch power capability.

        Args:
            http_fn (func): A function to send GET and PUT http commands
            ip_address (str): The IP address of the Powerswitch.
            command_dict (dict): A dictionary containing the command used for each method.
                                 The dictionary must contain the following keys:
                                 - ADJUST_PORTS_MODE

            headers_dict (dict): A dictionary containg the required headers for the HTTP requests.
                                 - GET_PROP
                                 - SET_PROP

            device_name (str): name of the device this capability is attached to.
            total_ports (int): number of ports on device
        """

        super().__init__(device_name=device_name)
        self._http_fn = http_fn
        self._ip_address = ip_address
        self._command_dict = command_dict
        self._headers_dict = headers_dict
        self._total_ports = total_ports

        self.validate_required_keys(self._REQUIRED_COMMANDS, command_dict, "command_dict")
        self.validate_required_keys(self._REQUIRED_HEADERS, headers_dict, "headers_dict")

    @decorators.PersistentProperty
    def supported_modes(self):
        """Get the Powerswitch power modes supported by the dli_powerswitch."""
        return [OFF, ON]

    def get_all_ports_mode(self):
        """Get mode of all the device ports.

        Returns:
            list: a list of the all the ports mode.
        """
        response = self._http_fn("GET",
                                 self._command_dict["ADJUST_PORTS_MODE"].format(
                                     "all;", ip=self._ip_address),
                                 headers=self._headers_dict["GET_PROP"])
        ports_value = response.split(",")
        ports_mode = [ON if port == "true"
                      else OFF for port in ports_value]
        return ports_mode

    def get_mode(self, port):
        """Get mode of the specified port.

        Args:
            port (int): device port number

        Raises:
            GazooDeviceError: if port is None, port < 0, or port >= total_ports

        Returns:
            str: mode of specified port, either 'on' or 'off'.
        """
        self._validate_port("get_mode", port)
        response = self._http_fn("GET",
                                 self._command_dict["ADJUST_PORTS_MODE"].format(
                                     "=" + str(port), ip=self._ip_address),
                                 headers=self._headers_dict["GET_PROP"])
        if response == 'true':
            return ON
        else:
            return OFF

    @decorators.CapabilityLogDecorator(logger)
    def power_on(self, port):
        """Powers on the specified port.

        Args:
            port (int): device port number

        Raises:
            GazooDeviceError: if port is None, port < 0, or port >= total_ports
        """
        self._validate_port("power_on", port)
        logger.debug("{} Powering on powerswitch port {}".format(self._device_name, port))
        self._http_fn("POST",
                      self._command_dict["ADJUST_PORTS_MODE"].format(
                          "=" + str(port), ip=self._ip_address),
                      headers=self._headers_dict["SET_PROP"],
                      data={"value": "true"})

    @decorators.CapabilityLogDecorator(logger)
    def power_off(self, port):
        """Powers off the specified port.

        Args:
            port (int): device port number

        Raises:
            GazooDeviceError: if port is None, port < 0, or port >= total_ports
        """
        self._validate_port("power_off", port)
        logger.debug("{} Powering off powerswitch port {}".format(self._device_name, port))
        self._http_fn("POST",
                      self._command_dict["ADJUST_PORTS_MODE"].format(
                          "=" + str(port), ip=self._ip_address),
                      headers=self._headers_dict["SET_PROP"],
                      data={"value": "false"})

    @decorators.CapabilityLogDecorator(logger)
    def set_mode(self, mode, port):
        """Sets the given Powerswitch port to the mode specified.

        Args:
            mode (str): mode to set Powerswitch ports to.
                        e.g. 'off', 'on'
            port (int): the port to set.

        Raises:
          GazooDeviceError: invalid mode.
        """
        self._validate_mode(mode)
        if mode == ON:
            self.power_on(port=port)
        else:
            self.power_off(port=port)

    @decorators.CapabilityLogDecorator(logger)
    def set_all_ports_mode(self, mode):
        """Set all device ports to the specified mode.

        Args:
            mode (str): Mode to set all Powerswitch ports to. Valid modes are "on" or "off".

        Raises:
            GazooDeviceError: if input mode is not either 'on' or 'off'
        """
        self._validate_mode(mode)
        if mode == ON:
            data_value = "true"
        else:
            data_value = "false"

        self._http_fn("POST",
                      self._command_dict["ADJUST_PORTS_MODE"].format("all;", ip=self._ip_address),
                      headers=self._headers_dict["SET_PROP"],
                      data={"value": data_value})

    def _validate_mode(self, mode):
        """Verify mode given resides in the valid mode list.

        Args:
            mode (str): mode e.g. 'off', 'on'

        Raises:
            GazooDeviceError: if mode given is not found in the valid mode list.
        """
        if mode not in list(self.supported_modes):
            raise errors.GazooDeviceError("Mode {} not found in supported modes {!r}".
                                          format(mode,
                                                 self.supported_modes))

    def _validate_port(self, method_name, port):
        """Ensure port is a valid port number.

        Args:
            method_name(str): name of the method where _validate_port is being called
            port (int): device port number

        Raises:
             GazooDeviceError: if input port is not valid
        """
        if port is None or port <= -1 or port >= self._total_ports:
            raise errors.GazooDeviceError("Device {} {} failed. "
                                          "Port {} is invalid.".
                                          format(self._device_name, method_name, port))



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""J-Link transport class."""
import intelhex
import os.path
import six
from gazoo_device import errors
from gazoo_device.switchboard import transport_base
from pylink import jlink
from pylink import enums

DEFAULT_CHIP_NAME = "NRF52840_XXAA"


class JLinkTransport(transport_base.TransportBase):
    """J-Link transport to a device."""

    def __init__(self,
                 comms_address,
                 chip_name=DEFAULT_CHIP_NAME,
                 speed="auto",
                 jlink_interface=enums.JLinkInterfaces.SWD,
                 auto_reopen=True,
                 open_on_start=True):
        """Initialize the J-Link transport.

        Args:
            comms_address (str): serial number of the J-Link adapter.
            chip_name (str): usually the name of the target chip to connect to.
            speed (object): connection speed, int or str.
                            Valid values: [5-12000], "auto", "adaptive".
            jlink_interface (int): interface to use. See pylink.enums.JLinkInterfaces.
            auto_reopen (bool): flag indicating transport should be reopened if
                                unexpectedly closed.
            open_on_start (bool): flag indicating transport should be open on
                                  TransportProcess start.

        Raises:
            GazooDeviceError: if the J-Link DLL (comes with the J-Link SDK) is missing.
            TypeError: J-Link instantiation failed.

        Note:
            due to license restrictions, J-Link DLL cannot be packaged into GDM.
        """
        super(JLinkTransport, self).__init__(auto_reopen=auto_reopen, open_on_start=open_on_start)
        self.comms_address = comms_address
        self._chip_name = chip_name
        self._speed = speed
        self._jlink_interface = jlink_interface

        try:
            self._jlink = jlink.JLink()
        except TypeError as e:
            if "Expected to be given a valid DLL." in str(e):
                raise errors.GazooDeviceError(
                    "No J-Link DLL found. Install the J-Link SDK from "
                    "https://www.segger.com/downloads/jlink/#J-LinkSoftwareAndDocumentationPack. "
                    "Error: {!r}.".format(e))
            else:
                raise

    def close(self):
        """Close the J-Link transport."""
        if hasattr(self, "_jlink") and self._jlink.opened():
            self._stop_rtt_logging()
            self._jlink.close()

    def flash(self, image_path):
        """Flash the hex image onto the device.

        Args:
            image_path (str): path to the image file to be flashed onto the device.
        """
        self._stop_rtt_logging()
        self._jlink.halt()

        image = intelhex.IntelHex(os.path.abspath(image_path))
        for segment_start, segment_end in image.segments():
            segment_size = segment_end - segment_start
            segment = image.tobinarray(start=segment_start, size=segment_size)
            self._jlink.flash_write8(segment_start, segment)

        self.reset()
        self._start_rtt_logging()

    def is_open(self):
        """Returns True if the J-Link library is open and transport is connected to the target.

        Returns:
            bool: True if transport is open, False otherwise.
        """
        return self._jlink.opened() and self._jlink.connected() and self._jlink.target_connected()

    def reset(self):
        """Reset the board by toggling the reset pin and restarting the CPU."""
        self._jlink.reset()
        self._jlink.restart()

    def _close(self):
        """Override abstract _close hook."""
        pass

    def _open(self):
        """Open the J-Link DLL and connect to the target."""
        self._jlink.open(serial_no=self.comms_address)
        self._jlink.set_tif(self._jlink_interface)
        self._jlink.connect(chip_name=self._chip_name, speed=self._speed)
        self._start_rtt_logging()

    def _read(self, size=1024, timeout=None):
        """Reads up to size bytes from the device RTT buffer.

        Args:
            size (int): maximum number of bytes to read.
            timeout (float): noop.

        Returns:
            bytes: bytes read from transport.
        """
        data = self._jlink.rtt_read(0, size)  # returns a list of int values (bytes)
        return b"".join(six.int2byte(value) for value in data)

    def _start_rtt_logging(self):
        """Restart RTT logging."""
        self._jlink.rtt_start()

    def _stop_rtt_logging(self):
        """Stop RTT logging."""
        self._jlink.rtt_stop()

    def _write(self, data, timeout=None):
        """Writes the data provided into the device RTT buffer.

        Args:
            data (str): bytes to be written.
            timeout (float): noop.

        Returns:
            int: number of bytes written.
        """
        if isinstance(data, six.text_type):
            data = data.encode("utf-8", errors="replace")

        bytes_to_write = list(six.iterbytes(data))  # convert to list of int values
        return self._jlink.rtt_write(0, bytes_to_write)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Parser class for filtering log files in background.

Parser provides general purpose regular expression matching on a line by line basis with the
ability to extract both the last and history of matching events.

Parser JSON filter files are typically loaded from two directories:
* The gazoo_device Python package path
* A relative path from the current directory

When the Parser class is created it attempts to load the default JSON filter files from the
following two locations:
* <path_to_gazoo_device_package>/filters/common
* <path_to_gazoo_device_package>/filters/<device_type>

Then the Parser class attempts to load each file or directory specified in the filters argument.

JSON filter file format (version 1.0):
* A single JSON object with the following keys:

    * "description" whose value is a string (optional)
    * "version" whose value is a JSON object with the following keys:

        * "major" whose value is an integer that specifies the major version of the file
        * "minor" whose value is an integer that specifies the major version of the file

    * "filters" whose value is an array of Filter JSON objects with the following keys:

        * "description" whose value is a string (optional)
        * "regex_match" whose value is a regular expression string
        * "name" whose value is a string

::

    {
        "description": "helpful optional description for all filters contained",
        "version": {"major": 1, "minor": 0},
        "filters": [
            {
                "description": "optional description for this specific filter",
                "regex_match": "required regular expression string",
                "name": "required_name"
            }
        ]
    }

As each JSON filter file is loaded and parsed each regular expression is assigned a name
consisting of the JSON filter file name (without the extension) plus the name assigned to each
Filter object (the "name" key) separated with a '.' (e.g. errors.assert). This is known in the
Parser code base as the "event_label" and is used as a lookup key to retrieve Parser filter
events.

The Parser class expects to store all matching events in an event file. To reduce the memory
footprint of the Parser class all matching events are immediately written to the event file by
the process_line method. The caller must provide the shared dynamic event dictionary, an event
file object, and the raw log line to this method. Keeping the shared dynamic event dictionary
and the opening and closing of the event file outside the Parser class helps to simplify the
Parser class logic and ensures that the file is closed by the caller as needed (e.g. exception
handling).

For each regular expression that matches the raw log line provided to the process_line method
the shared dynamic event dictionary is updated. When all regular expressions have been processed
an event JSON object containing the results of all matches by "event_lable" is written (and
flushed) to the event file. The event JSON object format is shown as follows:

JSON event file format (version 1.0):
* A single JSON object with the following keys:

    * "raw_log_line" whose value is the raw log line string parsed
    * "system_timestamp" whose value is a string containing the timestamp portion of the
      raw log line
    * "event_label" for each regular expression that matched the log line and whose value is an
      array of regular expression group data extracted

::

    {
        "system_timestamp": "2018-01-30 12:05:59.118746",
        "state.shutdown_reboot": ["The system is shutting down"],
        "raw_log_line": "<2018-01-30 12:05:59.118746> announce: The system is shutting down."
    }

Since event history can only be obtained from the event file, all event history commands require
the path to the event file to retrieve this information. Event history is obtained by using the
"tac" and "grep" unix tools to filter matching events in the event file and from Python extract
and decode each event JSON object as described above.
"""

from __future__ import absolute_import
from __future__ import division
import datetime
import json
import os
import re
import subprocess
import time

from gazoo_device import errors
from gazoo_device import gdm_logger

import six
from six.moves import filter

logger = gdm_logger.get_gdm_logger('parser')

FILTER_JSON_FORMAT_MAJOR_VERSION = 1
TIMESTAMP_FORMAT = "%Y-%m-%d %H:%M:%S.%f"


def _get_full_filter_name(name, filter_path):
    """Returns full filter name using information from name and filter_path provided.

    Args:
        name (str): as specified in filter dictionary in filter file
        filter_path (str): absolute path to filter file

    Returns:
        str: Full filter name to be used in Parser event file as event_label

    Note:
        The full filter name consists of the following elements:
            * filename without extension (e.g. crashes)
            * name as specified for each filter in filter file (e.g. segfault)

        Example:
            crashes.segfault
    """

    filename = os.path.basename(filter_path)
    file_only = os.path.splitext(filename)[0]
    filter_name = "{}.{}".format(file_only, name)
    return filter_name


def _get_datetime(dt_value):
    """Convert string to datetime object.

    Args:
        dt_value (str): string to convert.

    Raises:
        RuntimeError: if datetime string formatting fails to convert the value

    Returns:
        datetime: object converted from value

    Note:
      Looks for 1 or 3 formats:
         "%Y-%m-%d %H:%M:%S.%f %z"
         "%Y-%m-%d %H:%M:%S.%f"
         "%Y-%m-%d %H:%M:%S"
    """
    result = None
    if result is None:
        result = _get_datetime_from_format(dt_value, "%Y-%m-%d %H:%M:%S.%f %z")
    if result is None:
        result = _get_datetime_from_format(dt_value, "%Y-%m-%d %H:%M:%S.%f")
    if result is None:
        result = _get_datetime_from_format(dt_value, "%Y-%m-%d %H:%M:%S")
    if result is None:
        raise RuntimeError("Failed to convert '{}' into datetime object".
                           format(dt_value))
    return result


def _get_datetime_from_format(dt_value, dt_format):
    """Convert value provided into datetime object or None if not successful.

    The purpose of this method is to attempt to convert the value provided
    using the datetime format string also provided into a datetime object. If
    the value provided doesn't correspond to the datetime format provided
    then the datetime strptime method will throw a ValueError that this method
    will catch and consume by returning None for the result. This reduces the
    error checking the calling function must perform when testing multiple
    datetime formats that may or may not successfully convert the value
    provided into a datetime object.

    Args:
        dt_value (str): the datetime string value to be converted to a datetime object
        dt_format (str): the datetime format to use in conversion process

    Returns:
        datetime: if value can be formatted with dt_format provided into datetime object or None
        if unsuccessful.
    """

    result = None
    try:
        result = datetime.datetime.strptime(dt_value, dt_format)
    except ValueError:
        pass
    return result


def _get_last_event(device_event_file_path, event_label, timeout=1.0):
    """Returns event data matching the event label specified or None.

    Args:
        device_event_file_path (str): path to file with filtered event information.
        event_label (str): event label to lookup (e.g. 'x.state'). If None, return the
            last event regardless of event label.
        timeout (float): Timeout value in seconds. Example: 1.5.

    Raises:
        ValueError: if unable to read json data found in event_file_path.

    Returns:
        tuple: (dict, bool) where the first item is a dict which contains result data,
        and the second item is True if the command timed out, False otherwise. The result
        dict contains data in the following format::

            {"system_timestamp": datetime,
             "raw_log_line": str,
             event_label: list,
             ...}

        If event_label is provided, only the matching event label will be included;
        otherwise, all labels recorded in the event file will be included.
    """

    result = None
    timedout = False

    file_exists, remaining_timeout = _wait_for_event_file(device_event_file_path, timeout)
    if not file_exists:
        timedout = True
        return result, timedout

    timeout_str = "{:f}".format(remaining_timeout)

    if event_label is None:
        tail_cmd = ["timeout", timeout_str, "tail", "-n", "1", device_event_file_path]
        out = ""
        try:
            out = subprocess.check_output(tail_cmd).decode("utf-8", "replace")
        except subprocess.CalledProcessError as err:
            if err.returncode == 124:
                timedout = True
        events = _get_events_from_json_output([out], event_label)
    elif event_label:
        tac_cmd = ["tac", device_event_file_path]
        grep_cmd = ["timeout", timeout_str, "grep", "-m", "1", "-w", event_label]
        tac_proc = subprocess.Popen(tac_cmd, stdout=subprocess.PIPE)
        grep_proc = subprocess.Popen(grep_cmd, stdin=tac_proc.stdout, stdout=subprocess.PIPE)
        tac_proc.stdout.close()
        out, _ = grep_proc.communicate()
        if grep_proc.returncode == 124:
            timedout = True
        tac_proc.terminate()
        events = _get_events_from_json_output([out], [event_label])
    else:
        events = None

    if events:
        result = events[0]

    return result, timedout


def _get_events_from_json_output(json_events, event_labels=None):
    """Returns list of event data history matching event_labels provided.

    Args:
        json_events (list): of JSON object strings to return history for.
        event_labels (list): list of event labels to lookup (e.g. ['x.state', 'x.restored'])


    Raises:
        ValueError: when parsing JSON objects from json_events fails

    Returns:
        list: a list of dictionary objects for each event with the following format:
            {"system_timestamp": timestamp as a datetime object,
             "raw_log_line": raw log line as a string
             event_label: a list of regex group data matches, empty list if no groups
             ...}

        If multiple event labels match the same line, there will be one event_label
        in the dictionary for each match.
    """

    result = []
    for json_object in json_events:
        if json_object:
            try:
                event_dict = json.loads(json_object)
            except ValueError as err:
                logger.info("Failed to parse event log line; skipping. Err: {!r}".
                            format(err))
                continue
            if event_labels is not None:
                filtered_dict = {"system_timestamp": _get_datetime(event_dict["system_timestamp"]),
                                 "raw_log_line": event_dict["raw_log_line"]}
                for event_label in event_labels:
                    if event_label in event_dict:
                        filtered_dict[event_label] = event_dict[event_label]
                result.append(filtered_dict)
            else:
                event_dict["system_timestamp"] = _get_datetime(event_dict["system_timestamp"])
                result.append(event_dict)
    return result


def _get_all_event_history(device_event_file_path, event_labels, timeout=10.0):
    """Returns list of event history matching event labels specified.

    Args:
        device_event_file_path (str): path to file with filtered event information
        event_labels (list): list of event labels to lookup (e.g. ['x.state',
        'x.restored']). If None, return all events.
        timeout (float): Timeout value in seconds. Example: 10.0.

    Raises:
        ValueError: on (unlikely) invalid JSON data found in self.event_file_path file.

    Returns:
       tuple: (list of event objects (dict), bool whether or not command timedout )

    Note:
        The Parser standard is to return the list from newest to oldest, therefore we reverse
        the output returned by grep.
    """

    result = []
    timedout = False

    file_exists, remaining_timeout = _wait_for_event_file(device_event_file_path, timeout)
    if not file_exists:
        timedout = True
        return result, timedout

    timeout_str = "{:f}".format(remaining_timeout)

    if event_labels is None:
        tac_cmd = ["timeout", timeout_str, "tac", device_event_file_path]
        out = ""
        try:
            out = subprocess.check_output(tac_cmd).decode("utf-8", "replace")
        except subprocess.CalledProcessError as err:
            if err.returncode == 124:
                timedout = True
        json_events = out.splitlines()
    else:
        grep_cmd = ["timeout", timeout_str, "grep", "-w"]
        for event_label in event_labels:
            if event_label:
                grep_cmd.append("-e")
                grep_cmd.append(event_label)
        grep_cmd.append(device_event_file_path)
        grep_proc = subprocess.Popen(grep_cmd, stdout=subprocess.PIPE)
        out, _ = grep_proc.communicate()
        if grep_proc.returncode == 124:
            timedout = True
        json_events = out.splitlines()
        json_events.reverse()

    return _get_events_from_json_output(json_events, event_labels), timedout


def _get_event_history_count(device_event_file_path, event_label, timeout=10.0):
    """Returns count of event history matching event label specified.

    Args:
        device_event_file_path (str): to extract event history from
        event_label (str): event label to lookup (e.g. 'x.state')
        timeout (float): Timeout value in seconds. Example: 10.0.

    Raises:
        ValueError: on (unlikely) invalid JSON data found in self.event_file_path file.

    Returns:
        tuple: (int count of all events found, bool of whether command timed out).
    """

    result = 0
    timedout = False

    file_exists, remaining_timeout = _wait_for_event_file(device_event_file_path, timeout)
    if not file_exists:
        timedout = True
        return result, timedout

    timeout_str = "{:f}".format(remaining_timeout)

    grep_cmd = [
        "timeout",
        timeout_str,
        "grep",
        "-c",
        "-w",
        event_label,
        device_event_file_path]
    grep_proc = subprocess.Popen(grep_cmd, stdout=subprocess.PIPE)
    out, _ = grep_proc.communicate()
    if grep_proc.returncode == 124:
        timedout = True

    if out:
        result = int(out.strip())

    return result, timedout


def _get_limited_event_history(device_event_file_path, event_labels, limit, timeout=10.0):
    """Returns list of event history matching event labels specified.

    Args:
        device_event_file_path(str): path to file with filtered event information
        event_labels(list): list of event labels to lookup(e.g. ['x.state',
            'x.restored']). If None, return all events up to limit.
        limit(int): the number of event data elements returned
        timeout(float): Timeout value in seconds. Example: 10.0.

    Raises:
        ValueError: on(unlikely) invalid JSON data found in self.event_file_path file

    Returns:
        tuple: (list of event dicts up to 'limit' length, bool whether or not timed out)

    Note:
        Event info listed in most recent to least recent order. If regex_match for event_label
        has groups, it will return the group match, else it will return the raw log line.
        Implemented via grep and tac for time efficiency.
    """

    result = []
    timedout = False

    file_exists, remaining_timeout = _wait_for_event_file(device_event_file_path, timeout)
    if not file_exists:
        timedout = True
        return result, timedout

    timeout_str = "{:f}".format(remaining_timeout)

    if event_labels is None:
        tac_cmd = ["tac", device_event_file_path]
        head_cmd = ["timeout", timeout_str, "head", "-n", str(limit)]
        tac_proc = subprocess.Popen(tac_cmd, stdout=subprocess.PIPE)
        head_proc = subprocess.Popen(head_cmd, stdin=tac_proc.stdout, stdout=subprocess.PIPE)
        tac_proc.stdout.close()
        out, _ = head_proc.communicate()
        if head_proc.returncode == 124:
            timedout = True
        tac_proc.terminate()
        json_events = out.splitlines()
    else:
        tac_cmd = ["tac", device_event_file_path]
        grep_cmd = ["timeout", timeout_str, "grep", "-m", str(limit), "-w"]
        for event_label in event_labels:
            if event_label:
                grep_cmd.append("-e")
                grep_cmd.append(event_label)
        tac_proc = subprocess.Popen(tac_cmd, stdout=subprocess.PIPE)
        grep_proc = subprocess.Popen(grep_cmd, stdin=tac_proc.stdout, stdout=subprocess.PIPE)
        tac_proc.stdout.close()
        out, _ = grep_proc.communicate()
        if grep_proc.returncode == 124:
            timedout = True
        tac_proc.terminate()
        json_events = out.splitlines()

    return _get_events_from_json_output(json_events, event_labels), timedout


def _wait_for_event_file(event_file_path, timeout):
    """Wait up to the timeout period for the event file to be created.

    Args:
        event_file_path(str): Full path to the event file.
        timeout(float): Timeout value in seconds. Example: 1.5.

    Returns:
        tuple: (bool, float) A tuple where the first value is True if the file was found before
            the timeout expired, and False otherwise, and the second value is the amount
            of timeout time remaining.  For example, (True, 0.958) would indicate the file
            was found with 0.958 seconds of timeout left, and (False, 0) indicates that
            the event file was not found.  If the file is not found, the remaining time
            will always be 0.
    """

    start_time = time.time()
    end_time = start_time + timeout

    while True:
        timeout_remaining = end_time - time.time()
        if timeout_remaining <= 0:
            return (False, 0)
        elif os.path.exists(event_file_path):
            return (True, timeout_remaining)
        else:
            time.sleep(0.1)


class ParserResult(object):
    """Store the results of a Parser API call.

    Attributes:
        timedout (bool): True if the Parser request timed out; False otherwise
        results_list (list): A list of Event Results
        count (int): The number of Event Results returned

    Notes:
        An Event Result is a dict with the following format:
            {"system_timestamp": datetime,
             "raw_log_line": str,
             event_label: list,
             ...}

        The value of the "system_timestamp" key is a datetime object with the timestamp
        when the log line was captured by GDM. The value for the "raw_log_line" key is a string
        containing the entire log line captured by GDM, not including the timestamp added
        by GDM.

        The Event Result also contains one or more key - value pairs representing event
        matches for that log line, i.e., there will always be at least one event match,
        but there may be more than one if multiple event regexes match the same line.  For
        each match, the key is the event_label, and the value is a list of group matches
        for the corresponding regex. If no groups were specified in the event regex, the
        group matches list will be empty.
    """

    def __init__(self, timedout, results_list, count):
        self.timedout = timedout
        self.results_list = results_list
        self.count = count


class _EventMatch(object):
    """Encapsulates matching events with event time delta.
    """

    def __init__(self,
                 event_delta,
                 event_list):
        """Initializes _EventMatch.

        Args:
            event_delta (timedelta): Difference between event datetime instances
            event_list (list): Containing filter events

        Note:
            Filter events in event_list typically appear as::

                [
                {"basic.reboot_trigger": [],
                "system_timestamp": datetime.datetime.strptime("2018-06-18 15:13:48.011747",
                                                               "%Y-%m-%d %H:%M:%S.%f"),
                "raw_log_line": "Note: GDM triggered reboot"},

                {"basic.bootup": [],
                "system_timestamp": datetime.datetime.strptime("2018-06-18 15:13:49.092755",
                                                               "%Y-%m-%d %H:%M:%S.%f"),
                "raw_log_line": "[0:00:00.035] [APPL] Welcome to Lock App"}
                ]
        """

        self.event_delta = event_delta
        self.event_list = event_list


class Parser(object):
    """Parser class for filtering log lines."""

    def __init__(self, filters, event_file_path="unknown.txt", device_name="unknown"):
        self._filters_dict = {}
        self.name = device_name
        self.event_file_path = event_file_path
        self.load_filters(filters)

    def get_event_history(self, event_labels=None, count=None, timeout=10.0):
        r"""Returns up to count elements of event data matching given list of event labels.

        Args:
            event_labels (list): list of event labels to lookup(example:
                ["power.state", "power.lost"]). If None(default), return all events.
            count (int): of event data elements to return (Default: None for all events).
            timeout (float): Timeout value in seconds. Example: 10.0.

        Raises:
            GazooParserError: on error parsing event file.

        Returns:
            ParserResult: An object containing events matching any of the event labels
            provided in event_labels, or all available events if event_labels in
            None. The event data is stored in the .results_list attribute and the count of
            results in the .count attribute. If no matching events are found, the
            .results_list attribute will be set to the empty list.

            If the .timedout attribute is True, the operation timed out before all events
            were searched, or the event file was not found before the timeout expired.

        Notes:
            Event info is listed in most recent to least recent order.

            The following examples use a hypothetical power.json filter file with the
            following filters:

            .. code-block:: none

                {
                    "name": "state",
                    "regex_match": "power:(\\w+)"
                },
                {
                    "name": "lost",
                    "regex_match": "power lost"
                },
                {
                    "name": "restored",
                    "regex_match": "power restored"
                }

            The original log file being read looks like this:

            .. code-block:: none

                <2018-02-02 10:32:07.167234> [APPL] Spoke: power restored, powering on
                <2018-02-02 12:00:57.154328> [APPL] Spoke: power lost, powering down
                <2018-02-02 12:00:57.154328> power:OFF
                <2018-02-02 12:01:22.122824> [APPL] Spoke: power restored, powering on
                <2018-02-02 12:01:22.122824> power:ON
                <2018-02-02 12:02:18.121948> [APPL] Spoke: power lost, powering down
                <2018-02-02 12:02:18.121948> power:OF

            Example output with event_labels ["power.state"], showing group matches:

            .. code-block:: none

                result.timedout = False
                result.results_list =
                    [{'system_timestamp': datetime.datetime(2018, 2, 2, 12, 2, 18, 121948),
                      'raw_log_line': 'power:OFF',
                      'power.state': ['OFF']},
                     {'system_timestamp': datetime.datetime(2018, 2, 2, 12, 1, 22, 122824),
                      'raw_log_line': 'power:ON',
                      'power.state': ['ON']},
                     {'system_timestamp': datetime.datetime(2018, 2, 2, 12, 0, 57, 154328),
                      'raw_log_line': 'power:OFF',
                      'power.state': ['OFF']}]
                result.count = 3

            Example output with event_labels ["power.lost"], showing matches with no groups:

            .. code-block:: none

                result.timedout = False
                result.results_list =
                    [{'system_timestamp': datetime.datetime(2018, 2, 2, 12, 2, 18, 121948),
                      'raw_log_line: '[APPL] Spoke: power lost, powering down',
                      'power.lost': []},
                     {'system_timestamp': datetime.datetime(2018, 2, 2, 12, 0, 57, 154328),
                      'raw_log_line': '[APPL] Spoke: power lost, powering down',
                      'power.lost': []}]
                result.count = 2

            Example output with event_labels["power.restored"], showing matches with no
            groups and count=1:

            .. code-block:: none

                result.timedout = False
                result.results_list =
                    [{'system_timestamp': datetime.datetime(2018, 2, 2, 12, 1, 22, 122824),
                      'raw_log_line': '[APPL] Spoke: power restored, powering on',
                      'power.restored': []}
                result.count = 1

            Example output with event_labels ["power.lost", "power.restored"]:

            .. code-block:: none

                result.timedout = False
                result.results_list =
                    [{'system_timestamp': datetime.datetime(2018, 2, 2, 12, 2, 18, 121948),
                      'raw_log_line: '[APPL] Spoke: power lost, powering down',
                      'power.lost': []},
                     {'system_timestamp': datetime.datetime(2018, 2, 2, 12, 1, 22, 122824),
                      'raw_log_line': '[APPL] Spoke: power restored, powering on',
                      'power.restored': []},
                     {'system_timestamp': datetime.datetime(2018, 2, 2, 12, 0, 57, 154328),
                      'raw_log_line': '[APPL] Spoke: power lost, powering down',
                      'power.lost': []}
                     {'system_timestamp': datetime.datetime(2018, 2, 2, 10, 32, 7, 167234),
                      'raw_log_line': '[APPL] Spoke: power restored, powering on',
                      'power.restored': []}]
                result.count = 4
        """
        if event_labels:
            self.verify_event_labels(event_labels,
                                     error_message="{} get_event_history failed.".format(
                                         self.name))
        timedout = False
        try:
            if count:
                history_results, timedout = _get_limited_event_history(self.event_file_path,
                                                                       event_labels,
                                                                       count,
                                                                       timeout=timeout)
            else:
                history_results, timedout = _get_all_event_history(self.event_file_path,
                                                                   event_labels,
                                                                   timeout=timeout)
            return ParserResult(timedout=timedout,
                                results_list=history_results,
                                count=len(history_results))
        except Exception as err:
            raise errors.GazooParserError(
                "Retrieving events {} history from {} failed. Error {!r}".
                format(event_labels, self.event_file_path, err))

    def get_event_history_count(self, event_label, timeout=10.0):
        """Returns count of event history matching event label specified.

        Args:
            event_label(str): event label to lookup(e.g. power.state)
            timeout(float): Timeout value in seconds. Example: 10.0.

        Raises:
            GazooParserError: on error parsing event file

        Returns:
            ParserResult: An object containing the count of results in the .count
            attribute. The .results_list field is set to the empty list.

            If the .timedout attribute is True, the operation timed out before all events
            were searched, or the event file was not found before the timeout expired.
        """
        self.verify_event_labels([event_label],
                                 error_message="%s get_event_history_count failed." %
                                 self.name)

        try:
            count, timedout = _get_event_history_count(self.event_file_path,
                                                       event_label,
                                                       timeout=timeout)
            return ParserResult(timedout=timedout, results_list=[], count=count)
        except Exception as err:
            raise errors.GazooParserError("Retrieving event {} history from {} failed. Error {!r}".
                                          format(event_label, self.event_file_path, err))

    def get_event_label_dict(self, pattern=None):
        """Takes an event label pattern, and returns a dictionary of matching event labels.

        Args:
            pattern(str): Regex for event labels. Default: None

        Returns:
            dict: containing keys(i.e. event labels), and values(i.e regular expressions)

        Raises:
            GazooParserError: if pattern provided is an invalid regular expression

        Note:
            The pattern regular expression provided will be matched strictly against
            the beginning of each event_label.
            If no pattern is provided, the _filters_dict is returned.
        """

        if pattern is None:
            return {key: value.pattern for key, value in six.iteritems(self._filters_dict)}
        else:
            try:
                result_dict = {}
                regex = re.compile(pattern)
                for event_label in list(filter(regex.match, list(self._filters_dict))):
                    result_dict[event_label] = self._filters_dict[event_label].pattern
                return result_dict
            except Exception as err:
                raise errors.GazooParserError("Compiling regular expression pattern {} failed. "
                                              "Error: {!r}".
                                              format(pattern, err))

    def get_event_labels(self, pattern=None):
        """Takes an event label pattern, and returns a list of matching event labels.

        Args:
            pattern(str):  Regex for event labels. Default: None

        Returns:
            list: of matching event label strings

        Raises:
            GazooParserError: if pattern provided is an invalid regular expression

        Note:
            The pattern regular expression provided will be matched strictly against
            the beginning of each event_label.
            If no pattern is provided, all filter event labels are returned.
        """

        if pattern is None:
            return list(self._filters_dict.keys())
        else:
            try:
                regex = re.compile(pattern)
                return list(filter(regex.match, list(self._filters_dict.keys())))
            except Exception as err:
                raise errors.GazooParserError("Compiling regular expression pattern {} failed. "
                                              "Error {!r}".
                                              format(pattern, err))

    def get_last_event(self, event_labels=None, timeout=1.0):
        r"""Returns the most recent matching event for each item in the list of event labels.

        Args:
            event_labels (list): list of event labels to lookup (example: ["power.state",
                "power.lost"]). None (default) returns the most recent event regardless of
                label.
            timeout (float): Timeout value in seconds. Example: 1.5.

        Raises:
            GazooParserError: on error parsing event file.
            GazooDeviceError: Label doesn't exist or
                             event_labels is not type list

        Returns:
            ParserResult: An object containing the most recent event match for each event
            label in the event_labels list, or the most recent event if
            event_labels is None. The event data is stored in the .results_list
            attribute, and the count of results is in the .count attribute. If no matching
            events are found, the .results_list attribute will be set to the empty
            list.

            If the .timedout attribute is True, the operation timed out before all events
            were searched, or the event file was not found before the timeout expired.

        Notes:
            The event_label is the desired filter filename without extension a . and the filter
            name (e.g. power.state).

            The following examples use a hypothetical power.json filter file with the
            following filters:

            .. code-block:: none

                {
                    "name": "state",
                    "regex_match": "power:(\\w+)"
                },
                {
                    "name": "lost",
                    "regex_match": "power lost"
                },
                {
                    "name": "restored",
                    "regex_match": "power restored"
                }

            The original log file being read looks like this:

            .. code-block:: none

                <2018-02-02 10:32:07.167234> [APPL] Spoke: power restored, powering on
                <2018-02-02 12:00:57.154328> [APPL] Spoke: power lost, powering down
                <2018-02-02 12:00:57.154328> power:OFF
                <2018-02-02 12:01:22.122824> [APPL] Spoke: power restored, powering on
                <2018-02-02 12:01:22.122824> power:ON
                <2018-02-02 12:02:18.121948> [APPL] Spoke: power lost, powering down
                <2018-02-02 12:02:18.121948> power:OFF

            Example output with event_labels=["power.state"], showing group matches:

            .. code-block:: none

                result.timedout = False
                result.results_list =
                    [{'system_timestamp': datetime.datetime(2018, 2, 2, 12, 2, 18, 121948),
                      'raw_log_line': 'power:OFF',
                      'power.state': ['OFF']}]
                result.count = 1

            Example output with event_labels=["power.lost"], showing matches with no groups:

            .. code-block:: none

                result.timedout = False
                result.results_list =
                    [{'system_timestamp': datetime.datetime(2018, 2, 2, 12, 2, 18, 121948),
                      'raw_log_line': '[APPL] Spoke: power lost',
                      'power.lost': []}]
                result.count = 1

            Example output with event_labels=["power.state", "power.lost", "power.restored"]:

            .. code-block:: none

                result.timedout = False
                result.results_list =
                    [{'system_timestamp': datetime(2018, 2, 2, 12, 2, 18, 121948),
                      'raw_log_line': 'power:OFF',
                      'power.state': ['OFF']},
                     {'system_timestamp': datetime(2018, 2, 2, 12, 2, 18, 121948),
                      'raw_log_line': '[APPL] Spoke: power lost',
                      'power.lost': []}
                     {'system_timestamp': datetime(2018, 2, 2, 12, 1, 22, 122824),
                      'raw_log_line': '[APPL] Spoke: power restored, powering on',
                      'power.restored': []}]
                results.count = 3
        """
        if event_labels is None:
            event_labels = [None]
        else:
            self.verify_event_labels(event_labels,
                                     error_message="%s get_last_event failed." %
                                     self.name)

        results = []
        any_timed_out = False

        try:
            for event_label in event_labels:
                event_data, timedout = _get_last_event(self.event_file_path,
                                                       event_label,
                                                       timeout)
                if event_data:
                    results.append(event_data)
                any_timed_out |= timedout

            return ParserResult(timedout=any_timed_out, results_list=results, count=len(results))

        except Exception as err:
            raise errors.GazooParserError("{} get_last_event failed for {}. Error: {!r}.".
                                          format(self.name, event_labels, err))

    def get_last_event_state(self, event_label, event_name=None,
                             raise_error=False, timeout=1.0, group_index=0):
        r"""Return the matching group of the event label regex if one is found.

        Args:
            event_label(str): The event label based on the filter name and filter file name.
            event_name (str): name of event. Defaults to event_label.
            raise_error(bool): Raise an exception if the event has not been seen or otherwise
                         return an empty string.
            timeout (float): time to parse events.
            group_index (int): index of the match group to return (default 0 for first).

        Returns:
            str: The result from the call to get_last_event and getting the first group.

        Raises:
            GazooParserError: if there is an error getting the last event.

        .. code-block:: none
                {
                    "name": "state",
                    "regex_match": "power:(\\w+)"
                }

            The original log file being read looks like this:

            .. code-block:: none

                <2018-02-02 10:32:07.167234> [APPL] Spoke: power restored, powering on
                <2018-02-02 12:00:57.154328> [APPL] Spoke: power lost, powering down
                <2018-02-02 12:00:57.154328> power:OFF
                <2018-02-02 12:01:22.122824> [APPL] Spoke: power restored, powering on
                <2018-02-02 12:01:22.122824> power:ON
                <2018-02-02 12:02:18.121948> [APPL] Spoke: power lost, powering down
                <2018-02-02 12:02:18.121948> power:OFF

            With event_labels=["power.state"], it would return 'OFF'
        """
        if not event_name:
            event_name = event_label
        result = self.get_last_event([event_label], timeout=timeout)

        if not result.results_list:
            msg = "{} get event state failed. {} not yet seen in logs".format(
                self.name, event_name)
        else:
            match_groups = result.results_list[0][event_label]
            if group_index < len(match_groups):
                return match_groups[group_index]
            else:
                msg = ("{} get event state failed. Group index ({}) is greater than number of "
                       "match groups ({})").format(self.name, group_index, match_groups)

        if raise_error:
            raise errors.GazooParserError(msg)
        else:
            logger.warning(msg)
            return ""

    def get_unexpected_reboots(self):
        """Device's event file is used to identify unmatched reboot events.

        Returns:
            list: unexpected reboot event

        Note:
            For example in the following case there are two reboot triggers but only one
            bootup event::

                {"basic.reboot_trigger": [],
                 "system_timestamp": datetime.datetime.strptime("2018-06-18 15:13:48.011747",
                                                                "%Y-%m-%d %H:%M:%S.%f"),
                 "raw_log_line": "Note: GDM triggered reboot"},
                {"basic.reboot_trigger": [],
                 "system_timestamp": datetime.datetime.strptime("2018-06-18 15:13:52.825231",
                                                                "%Y-%m-%d %H:%M:%S.%f"),
                 "raw_log_line": "Note: GDM triggered reboot"}
                {"basic.bootup": [],
                 "system_timestamp": datetime.datetime.strptime("2018-06-18 15:13:58.540417",
                                                                "%Y-%m-%d %H:%M:%S.%f"),
                 "raw_log_line": "[0:00:00.035] [APPL] Welcome to Lock App"}

            This method would return::

                [{"basic.reboot_trigger": [],
                 "system_timestamp": datetime.datetime.strptime("2018-06-18 15:13:48.011747",
                                                                "%Y-%m-%d %H:%M:%S.%f"),
                 "raw_log_line": "Note: GDM triggered reboot"}]
        """

        _, remaining_bootups = self.get_unmatched_events(event_cause_label="basic.reboot_trigger",
                                                         event_effect_label="basic.bootup")

        return remaining_bootups

    def get_unmatched_events(self,
                             event_cause_label="basic.reboot_trigger",
                             event_effect_label="basic.bootup"):
        """Device's event file is used to identify unmatched events.

        Args:
            event_cause_label (str): Name of event label causing event
            event_effect_label (str): Name of event label signifying effect event

        Returns:
            tuple:  list of unmatched cause event dict entries
                    list of unmatched effect event dict entries

        Note:
            For example in the following case there are two reboot triggers but only one
            bootup event::

                {"basic.reboot_trigger": [],
                 "system_timestamp": datetime.datetime.strptime("2018-06-18 15:13:48.011747",
                                                                "%Y-%m-%d %H:%M:%S.%f"),
                 "raw_log_line": "Note: GDM triggered reboot"},
                {"basic.reboot_trigger": [],
                 "system_timestamp": datetime.datetime.strptime("2018-06-18 15:13:52.825231",
                                                                "%Y-%m-%d %H:%M:%S.%f"),
                 "raw_log_line": "Note: GDM triggered reboot"}
                {"basic.bootup": [],
                 "system_timestamp": datetime.datetime.strptime("2018-06-18 15:13:58.540417",
                                                                "%Y-%m-%d %H:%M:%S.%f"),
                 "raw_log_line": "[0:00:00.035] [APPL] Welcome to Lock App"}

            This method would return the following tuple::

                ([{"basic.reboot_trigger": [],
                 "system_timestamp": datetime.datetime.strptime("2018-06-18 15:13:48.011747",
                                                                "%Y-%m-%d %H:%M:%S.%f"),
                 "raw_log_line": "Note: GDM triggered reboot"}],
                [])
        """
        event_cause_list = []  # Cause events matching event_cause_label
        event_match_list = []  # EventMatch objects holding events matched together based on time
        remaining_cause_list = []  # Cause events with no matching effect events
        remaining_effect_list = []  # Effect events with no matching cause events

        event_cause_hist = self.get_event_history([event_cause_label])
        if not event_cause_hist.timedout:
            remaining_cause_list = event_cause_hist.results_list[:]
            event_cause_list = event_cause_hist.results_list
        event_effect_hist = self.get_event_history([event_effect_label])
        if not event_effect_hist.timedout:
            remaining_effect_list = event_effect_hist.results_list[:]

        # List of each cause event's datetime.datetime value
        cause_datetime_list = [event["system_timestamp"] for event in event_cause_list]

        # A copy of cause_datetime_list adjusted during event match processing
        wrk_datetime_list = cause_datetime_list[:]

        # Create list of matching events
        for effect_event in event_effect_hist.results_list:
            effect_datetime = effect_event["system_timestamp"]
            try:
                nearest_cause_datetime = self._nearest_datetime(wrk_datetime_list,
                                                                effect_datetime)
            except errors.GazooParserError:  # Means wrk_datetime_list is empty
                break
            delta_time = effect_datetime - nearest_cause_datetime
            cause_event = event_cause_list[cause_datetime_list.index(nearest_cause_datetime)]
            event_match = _EventMatch(event_delta=delta_time,
                                      event_list=[cause_event, effect_event])
            event_match_list.append(event_match)
            # The datetime has been used, adjust list to avoid reuse
            del wrk_datetime_list[wrk_datetime_list.index(nearest_cause_datetime)]
        if len(event_match_list) > 1:
            event_match_list.sort(key=lambda match_obj: match_obj.event_delta)

        # Use list of matching events to determine unmatched events
        for event_match in event_match_list:
            if remaining_cause_list and remaining_effect_list:
                cause_event = event_match.event_list[0]
                effect_event = event_match.event_list[1]
                try:
                    idx = remaining_cause_list.index(cause_event)
                    del remaining_cause_list[idx]
                except ValueError:
                    pass
                try:
                    idx = remaining_effect_list.index(effect_event)
                    del remaining_effect_list[idx]
                except ValueError:
                    pass
        return remaining_cause_list, remaining_effect_list

    def load_filter_file(self, filter_path):
        """Load JSON filter file specified.

        Args:
            filter_path (str): path to JSON filter file

        Raises:
            GazooParserError: when loading JSON filter file fails or on unexpected JSON key values

        Note:
            Only files ending in .json are loaded, others are skipped

            Example:
                {
                    "description": "Device assert events",
                    "version": {"major": 1, "minor": 0},
                    "filters": [
                        {
                            "description": "match and extract assert message"
                            "regex_match": "ASSERT: (.+)",
                            "name": "state"
                        }
                    ]
                }
        """
        logger.debug("Adding filter file {}", filter_path)
        try:
            with open(filter_path, "r") as filter_file:
                try:
                    json_filter_data = json.load(filter_file)
                except Exception as err:
                    msg = "Unable to parse filter file {} as a json file. {!r}".format(
                        filter_path, err)
                    logger.debug(msg)
                    raise errors.GazooParserError(msg)
        except IOError:
            raise errors.GazooParserError("Unable to access filter path '{}'".format(filter_path))

        if "version" not in json_filter_data:
            raise errors.GazooParserError("Loading filter-file {} failed. Missing 'version' key.".
                                          format(filter_path))

        if "filters" not in json_filter_data:
            raise errors.GazooParserError("Loading filter-file {} failed. Missing 'filters' key.".
                                          format(filter_path))

        if not isinstance(json_filter_data["version"], dict):
            raise errors.GazooParserError("Loading filter-file {} failed. "
                                          "Expecting value of 'version' entry to be a dictionary "
                                          "but instead its a {}.".
                                          format(filter_path, type(json_filter_data["version"])))

        version_info = json_filter_data["version"]

        if "major" not in version_info:
            raise errors.GazooParserError("Loading filter-file {} failed. "
                                          "Missing 'major' key in 'version' value.".
                                          format(filter_path))

        if "minor" not in version_info:
            raise errors.GazooParserError("Loading filter-file {} failed. "
                                          "Missing 'minor' key in 'version' value.".
                                          format(filter_path))

        if not isinstance(version_info["major"], int):
            raise errors.GazooParserError("Loading filter-file {} failed. "
                                          "Expecting int for major version found {} instead.".
                                          format(filter_path, type(version_info["major"])))

        if not isinstance(version_info["minor"], int):
            raise errors.GazooParserError("Loading filter-file {} failed. "
                                          "Expecting int for minor version found {} instead.".
                                          format(filter_path, type(version_info["minor"])))

        if version_info["major"] != FILTER_JSON_FORMAT_MAJOR_VERSION:
            raise errors.GazooParserError("Loading filter-file {} failed. "
                                          "Found unexpected major version in JSON filter file.".
                                          format(filter_path))

        self._add_filters(json_filter_data["filters"], filter_path)

    def process_line(self, event_file, raw_log_line, header_length=29, log_filename=""):
        """Parser line against filters and write JSON result to event_file provided.

        Args:
            event_file (file): to write matching filter results to
            raw_log_line (str): to match filters against
            header_length (int): added by GDM to strip off from raw_log_line
            log_filename (str): name of log file raw_log_line came from

        Note:
            The expected format for the raw_log_line is one that contains the system timestamp
            between characters 1 and 27 as shown in the following example:

                <2018-02-01 14:39:00.492376> ...

            A line containing a event_entry should be written as follows:
                * Key: log_filename, Value: filename where raw_log_line was found
                * Key: matched_timestamp, Value: a host timestamp when match was found.
                * Key: raw_log_line, Value: the line provided without any trailing whitespace
                * Key: system_timestamp, Value: the timestamp portion of the raw log line
                * Key: <event_label>, Value: regular expression group tuple as an array which
                  might be empty

            In addition to writing results to the event file the <event_label> is used as a key
            to update the shared dynamic events dictionary with the same information to allow for
            easier access to the data for the last matching <event_label> in the main process
        """

        event_data = {}
        for filter_name, regex in self._filters_dict.items():
            match = regex.search(raw_log_line)
            if match:
                match_data = match.groups()
                event_data[filter_name] = match_data

        if event_data:
            if log_filename:
                event_data["log_filename"] = log_filename
            event_data["raw_log_line"] = raw_log_line.rstrip()[header_length:]
            event_data["system_timestamp"] = raw_log_line[1:27]
            event_data["matched_timestamp"] = \
                datetime.datetime.now().strftime(TIMESTAMP_FORMAT)
            event_file.write(json.dumps(event_data) + "\n")
            event_file.flush()

    def verify_event_labels(self, event_labels, error_message=""):
        """Verifies event_labels in correct format and exist.

        Args:
           event_labels (list): list of event_labels.
           error_message (str): error_message to use as header.

        Raises:
           GazooParserError: if bad format or certain labels don't exist.

        """
        if event_labels is not None and not isinstance(event_labels, list):
            raise errors.GazooParserError("{} Expecting event_labels to be a list "
                                          "or None but received {} instead.".
                                          format(error_message, type(event_labels)))
        if any(not isinstance(label, str) for label in event_labels):
            raise errors.GazooParserError("{} Expected event_labels to be a list of strings. "
                                          "Found {}.".format(error_message, event_labels))

        cur_event_labels = self.get_event_labels()
        for event_label in event_labels:
            if event_label not in cur_event_labels:
                raise errors.GazooParserError("{} label {} not available. Choose from: {}".format(
                    self.name, event_label, self.get_event_labels()))

    def wait_for_event_labels(self, event_labels, raise_error=False, timeout=20.0,
                              start_datetime=None):
        """Waits up to timeout seconds for event labels to appear in device logs.

        Args:
            event_labels (list): labels(strings) to wait for. Will wait for at least one of each.
            raise_error (bool): raise an error if the labels are not found in time.
            timeout (float): seconds to wait for labels to appear in logs.
            start_datetime (datetime): events before this time will be ignored.

        Raises:
            GazooParserError: if input format is bad
            GazooParserError: raise_error is True and not all labels were found.

        Returns:
            bool: whether all event labels were found.
        """
        self.verify_event_labels(event_labels,
                                 error_message="%s wait_for_event_labels failed." % self.name)
        logger.info("{} waiting up to {}s for the following labels: {}."
                    .format(self.name, timeout, event_labels))

        if start_datetime is None:
            start_datetime = datetime.datetime.now()
        start_time = time.time()
        end_time = start_time + timeout

        all_found = False

        while not all_found and time.time() < end_time:
            remaining_time = timeout - (time.time() - start_time)
            # len(event_labels) can be 0; avoid zero division
            time_per_label = remaining_time / max(len(event_labels), 1)

            any_timed_out = False
            found_labels = []
            missed_labels = []

            for event_label in event_labels:
                result = self.get_last_event([event_label], timeout=time_per_label)
                any_timed_out |= result.timedout
                if (result.results_list
                        and result.results_list[0]["system_timestamp"] >= start_datetime):
                    found_labels.append(event_label)
                else:
                    missed_labels.append(event_label)

            all_found = not any_timed_out and len(found_labels) == len(event_labels)

        if not all_found:
            msg = ("{}: not all events corresponding to labels were found in {}s. "
                   "Looked for labels: {}. Found labels: {}. Missed labels: {}."
                   .format(self.name, timeout, event_labels, found_labels, missed_labels))
            if raise_error:
                raise errors.GazooParserError(msg)
            else:
                logger.warning(msg)

        return all_found

    def _add_filter(self, filter_list, filter_path):
        """Adds filter dictionary expression from filter_path specified.

        Args:
            filter_list(list): of filter dictionaries definitions to add
            filter_path(str): that provided the list of filters being added

        Raises:
            GazooParserError: when one of the following occurs
                * filter dictionary is missing the name key
                * filter dictionary is missing the regex_match key
                * full filter name already exists(e.g. already loaded)
                * the regex_match is invalid(won't compile)
        """

        if "name" not in filter_list:
            raise errors.GazooParserError("Loading filter-file {} failed. "
                                          "Missing 'name' key in filter object '{!r}'.".
                                          format(filter_path, filter_list))
        if "regex_match" not in filter_list:
            raise errors.GazooParserError("Loading filter-file {} failed. "
                                          "Missing 'regex_match' key in filter object '{!r}'.".
                                          format(filter_path, filter_list))

        full_filter_name = _get_full_filter_name(filter_list["name"], filter_path)
        if full_filter_name in self._filters_dict:
            raise errors.GazooParserError("Loading filter-file {} failed. "
                                          "Filter named {} already exists.".
                                          format(filter_path, full_filter_name))

        try:
            self._filters_dict[full_filter_name] = re.compile(filter_list["regex_match"])
            logger.debug("Added filter {} from filter file {}", full_filter_name, filter_path)
        except re.error as err:
            raise errors.GazooParserError("Loading filter-file {} failed. Unable to"
                                          " compile regular expression value '{}'. Error {!r}".
                                          format(filter_path, filter_list["regex_match"], err))

    def _add_filters(self, filter_list, filter_path):
        """Adds filters loaded from file_name to list of filters to use.

        Args:
            filter_list(list): of filter dictionaries definitions to add
            filter_path(str): that provided the list of filters being added

        Raises:
            GazooParserError: when filters is not of type list
        """

        if not isinstance(filter_list, list):
            raise errors.GazooParserError("Loading filter-file {} failed. "
                                          "Expecting value of 'filter_list' entry to be a list "
                                          "but instead its a {}.".
                                          format(filter_path, type(filter_list)))

        for cur_filter in filter_list:
            self._add_filter(cur_filter, filter_path)

    def _load_filter_directory(self, filter_path):
        """Loads every .json file found in filter_path specified.

        Args:
            filter_path(str): path to filter directory

        Raises:
            GazooParserError: if unable to access the filter_path provided

        Note:
            Only files ending in .json are loaded, others are skipped(see load_filter_file
            method for an example)
        """

        try:
            logger.debug("Adding filters from directory {}", filter_path)
            for filter_file in os.listdir(filter_path):
                if filter_file.endswith(".json"):
                    filter_file_path = os.path.join(filter_path, filter_file)
                    self.load_filter_file(filter_file_path)
                else:
                    logger.debug("Skipping file {} missing .json extension", filter_path)
        except OSError:
            raise errors.GazooParserError("Unable to access filter path '{}'".
                                          format(filter_path))

    def load_filters(self, filters):
        """Load JSON filter files or directories specified.

        Args:
            filters(list): of JSON filter files or paths to filter files

        Raises:
            GazooParserError: if files or paths do not exist

        Note:
            Only files ending in .json are loaded, others are skipped(see load_filter_file
            method for an example)
        """
        if not filters:
            return
        if not isinstance(filters, list):
            raise errors.GazooParserError("Expecting 'filters' value to be a list "
                                          "but instead its a {}.".
                                          format(type(filters)))
        for filter_path in filters:
            if not os.path.exists(filter_path):
                raise errors.GazooParserError("Filter path '{}' doesn't exist".
                                              format(filter_path))
            elif os.path.isdir(filter_path):
                self._load_filter_directory(filter_path)
            else:
                self.load_filter_file(filter_path)

    def _nearest_datetime(self, datetime_list, target_datetime):
        """Locates a datetime.datetime closest to target datetime.

        Args:
            datetime_list(list): Containing datetime.datetime entries
            target_datetime(datetime.datetime): Desired target

        Returns:
            datetime.datetime: Closest entry to target datetime without exceeding
                               target datetime

        Raises:
            GazooParserError: If datetime_list is empty
        """

        if not datetime_list:
            raise errors.GazooParserError("Input parameter datetime_list length is zero. Required"
                                          " parameters: [datetime.datetime], datetime.datetime")
        work_list = [entry for entry in datetime_list if entry < target_datetime]
        if not work_list:
            raise errors.GazooParserError("work_list length is zero. Entries in datetime_list"
                                          " {} are not < target_datetime {}".
                                          format(datetime_list, target_datetime))
        return min(work_list,
                   key=lambda datetime_entry: abs(datetime_entry - target_datetime))



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Switch power capability interface.

This class defines the required API all flavors of the switch_power capability.
"""
import abc
from gazoo_device.capabilities.interfaces import capability_base


class SwitchPowerBase(capability_base.CapabilityBase):
    """Abstract base class defining the API for the switch_power capability."""

    @abc.abstractproperty
    def supported_modes(self):
        """Get auxiliary device supported power modes."""

    @abc.abstractmethod
    def power_on(self, port):
        """This command powers on the port specified.

        Args:
            port (int): Identifies which auxiliary device port to power on.
        """

    @abc.abstractmethod
    def power_off(self, port):
        """This command powers off the port specified.

        Args:
            port (int): Identifies which auxiliary device port to power off.
        """

    @abc.abstractmethod
    def set_mode(self, mode, port):
        """Sets the given auxiliary device port to the mode specified.

        Args:
            mode (str): auxiliary device mode to set.
            port (int): The port to set.

        Raises:
          GazooDeviceError: invalid port, or mode.
        """

    @abc.abstractmethod
    def set_all_ports_mode(self, mode):
        """Sets all auxiliary device ports to the mode specified.

        Args:
          mode (str): auxiliary device mode to set.

        Raises:
          GazooDeviceError: invalid mode.
        """

    @abc.abstractmethod
    def get_mode(self, port):
        """Gets the auxiliary device mode for the specified port.

        Args:
          port (int): Use this port to get the mode.

        Returns:
            str: auxiliary device port mode settings

        Raises:
          GazooDeviceError: invalid port.
        """

    @abc.abstractmethod
    def get_all_ports_mode(self):
        """Gets the auxiliary device mode for all ports.

        Returns:
          list: Returns a list of port modes where the index is port number.
        """



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base class for all auxiliary devices."""
import difflib
import functools
import inspect
import os
import re
import time
import weakref

from gazoo_device import config
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.base_classes import auxiliary_device_base
from gazoo_device.utility import common_utils
from gazoo_device.utility import supported_classes

logger = gdm_logger.get_gdm_logger()

ERROR_PREFIX = "Exception_"
TIMEOUTS = {"CONNECTED": 3,
            "SHELL": 60}


def get_log_filename(log_directory, device_name, name_prefix=""):
    """Returns the full path of log filename using the information provided.

    Args:
        log_directory (path): to where the log file should be created.
        device_name (str): to use in the log filename
        name_prefix (str): string to prepend to the start of the log file.

    Returns:
        str: Path to log filename using the information provided.
    """

    log_timestamp = time.strftime("%Y%m%d-%H%M%S")
    if name_prefix:
        log_file_name = "{}-{}-{}.txt".format(name_prefix, device_name, log_timestamp)
    else:
        log_file_name = "{}-{}.txt".format(device_name, log_timestamp)
    return os.path.join(log_directory, log_file_name)


class AuxiliaryDevice(auxiliary_device_base.AuxiliaryDeviceBase):
    """Interface containing required GDM APIs for first party devices."""

    def __init__(self,
                 manager,
                 device_config,
                 log_file_name=None,
                 log_directory=None):
        """Initialize a device class with the serial port information provided.

        Args:
            manager (gazoo_device.Manager): manager object for this instance of gazoo_device.
            device_config (dict): dict of two dicts, 'optional' device parameters and
                                  'persistent' device parameters.
            log_file_name (str): file name in the log directory for device logs.
            log_directory (str): directory in which the controller will create the log file.
        """
        self.manager_weakref = weakref.ref(manager)

        # Create a dictionary to store "properties".  For now keep the
        # classification of "persistent" and "optional".
        self.props = {'persistent_identifiers': device_config["persistent"],
                      'optional': device_config["options"]}

        self.device_parser = None

        self._regexes = {}
        self._commands = {}
        self._timeouts = TIMEOUTS.copy()
        # Initialize log files
        self.log_directory = log_directory
        if log_file_name:
            self._log_file_name = os.path.join(log_directory, log_file_name)
        else:
            log_name_prefix = device_config["log_name_prefix"]
            self._log_file_name = get_log_filename(log_directory,
                                                   self.name,
                                                   name_prefix=log_name_prefix)
        self.device_type = self.DEVICE_TYPE

    @decorators.SettableProperty
    def alias(self):
        """Returns the user-defined device alias (string)."""
        return self.props['optional']['alias']

    @decorators.PersistentProperty
    def communication_address(self):
        """Returns the name of the main communication port (e.g. ip address).

        Returns:
            str: path or address of main communication port.
        """
        return self.props['persistent_identifiers'].get('console_port_name')

    @decorators.PersistentProperty
    def commands(self):
        """Dictionary of commands issued to the device via shell."""
        return self._commands

    @decorators.DynamicProperty
    def connected(self):
        """Returns whether or not device is connected."""
        device_config = {'persistent': self.props["persistent_identifiers"]}
        return self.is_connected(device_config)

    @classmethod
    def get_dynamic_property_names(cls):
        """Returns a list of dynamic property names including capability ones."""
        return cls._get_property_names(decorators.DynamicProperty)

    @classmethod
    def get_persistent_property_names(cls):
        """Returns a list of persistent property names including capability ones."""
        names = cls._get_property_names(decorators.PersistentProperty)
        for property_value in config.CLASS_PROPERTY_TYPES:
            names += cls._get_property_names(property_value)
        return sorted(names)

    @classmethod
    def get_settable_property_names(cls):
        """Returns a list of settable property names including capability ones."""
        return cls._get_property_names(decorators.SettableProperty)

    @decorators.PersistentProperty
    def health_checks(self):
        """Returns list of methods to execute as health checks."""
        return [self.device_is_connected]

    @decorators.DynamicProperty
    def log_file_name(self):
        """Returns current log file name in use.

        Returns:
            str: Path to current log file name.
        """
        return self._log_file_name

    @decorators.PersistentProperty
    def regexes(self):
        """Regular expressions used to retrieve properties, events and states from device output.

        Returns:
            dict: mapping of name to regular expression.
        """
        return self._regexes

    @decorators.PersistentProperty
    def model(self):
        """Returns the device model."""
        return self.props['persistent_identifiers']['model']

    @decorators.PersistentProperty
    def name(self):
        return self.props['persistent_identifiers']['name']

    @decorators.PersistentProperty
    def serial_number(self):
        return self.props['persistent_identifiers']['serial_number']

    @decorators.PersistentProperty
    def timeouts(self):
        """Dictionary of default timeouts to use when expecting certain actions."""
        return self._timeouts

    @property
    def switchboard(self):
        """Instance for communicating with the device."""
        if not hasattr(self, "_switchboard"):
            switchboard_kwargs = self._COMMUNICATION_KWARGS.copy()
            switchboard_kwargs["communication_address"] = self.communication_address
            switchboard_kwargs["communication_type"] = self.COMMUNICATION_TYPE
            switchboard_kwargs["log_path"] = self.log_file_name
            switchboard_kwargs["device_name"] = self.name
            switchboard_kwargs["event_parser"] = self.device_parser
            self._switchboard = self.manager_weakref().create_switchboard(**switchboard_kwargs)

        return self._switchboard

    @decorators.health_check
    def check_create_switchboard(self):
        """Checks switchboard can be successfully created."""
        self.switchboard.add_log_note("{} switchboard successfully started.".format(self.name))

    @decorators.LogDecorator(logger)
    def check_device_ready(self):
        """Check device readiness by executing a series of health check methods."""
        self._execute_health_check_methods(self.health_checks)

    @decorators.LogDecorator(logger, decorators.DEBUG)
    def close(self):
        """Calls close on the communication ports and resets anything needed.

        Note:
            Resets the buttons and terminates the child processes.
        """
        if hasattr(self, '_switchboard'):
            self.switchboard.close()
            self.switchboard.ensure_path_unlocked(self.name, self.communication_address)
            del self._switchboard

        if hasattr(self, "manager_weakref"):
            manager_instance = self.manager_weakref()
            if manager_instance is not None:
                if self.name in manager_instance._open_devices:
                    self._log_object_lifecycle_event("close")
                    del manager_instance._open_devices[self.name]
            del manager_instance

    @decorators.health_check
    def device_is_connected(self):
        """Checks that device shows up as a connection on the host machine.

        Raises:
           DeviceNotConnectedError: if device is not connected.
        """
        device_config = {'persistent': self.props["persistent_identifiers"]}
        try:
            common_utils.retry(func=self.is_connected,
                               func_args=(device_config,),
                               is_successful=bool,
                               timeout=self.timeouts["CONNECTED"])
        except errors.CommunicationTimeoutError:
            raise errors.DeviceNotConnectedError(
                self.name, msg="device not reachable from host machine.")

    def get_dynamic_properties(self):
        """Returns a dictionary of prop, value for each dynamic property."""
        names = self.get_dynamic_property_names()
        return self._get_properties(names)

    def get_persistent_properties(self):
        """Returns a dictionary of prop, value for each persistent property."""
        names = self.get_persistent_property_names()
        return self._get_properties(names)

    def get_property(self, name, raise_error=False):
        """Retrieves a property value (can be nested).

        Args:
            name (str): name of a single property to fetch.
            raise_error (bool): raise error if unable to retrieve property

        Returns:
            object: value of the specified dynamic property.

        Raises:
            GazooDeviceError: property value is not a plain data object.
            AttributeError: property doesn't exist.
            Exception: property is not retrieveable for arbitrary reason.

        Note:
           Expects property value to be a plain data object.
           Returns a string when property doesn't exist or property raises an error.
           Able to process both device properties (firmware_version) and
           capability properties (wpan.ncp_address)
        """
        instance = self
        try:
            if '.' in name:  # capability property
                instance = getattr(self, name.split('.')[0])
                name = name.split('.')[1]

            value = getattr(instance, name)
            if callable(value):
                raise errors.GazooDeviceError("{}'s {} is a method".format(self.name, name))
            return value
        except AttributeError:
            if name in self.props['optional']:
                return self.props['optional'][name]
            if raise_error:
                raise
            close_matches = difflib.get_close_matches(name, self.get_property_names())
            return "{} does not have a known property '{}'. Close matches: {}".format(
                self.name,
                name,
                " or ".join(close_matches))
        except Exception as err:
            if raise_error:
                raise
            error_type = type(err).__name__
            logger.info("{} for {}, exception: {}".format(error_type, name, str(err)))
            return ERROR_PREFIX + error_type

    def get_property_names(self):
        """Returns a list of all property names."""
        full_list = list(self.get_dynamic_property_names())
        full_list.extend(self.get_persistent_property_names())
        full_list.extend(self.get_settable_property_names())
        full_list.extend(self.props['optional'])
        return list(set(full_list))

    def get_settable_properties(self):
        """Returns a dictionary of prop, value for each settable property."""
        names = self.get_settable_property_names()
        return self._get_properties(names)

    @classmethod
    def get_supported_capabilities(cls):
        """Returns a list of names of capabilities supported by this device class."""
        # Deduplicate names: there may be several flavors which share the same interface
        capability_names = {capability_class.get_capability_name()
                            for capability_class in cls.get_supported_capability_flavors()}
        return sorted(list(capability_names))

    @classmethod
    def get_supported_capability_flavors(cls):
        """Returns a set of all capability flavor classes supported by this device class."""
        capability_classes = [member.capability_classes for _, member in inspect.getmembers(cls)
                              if isinstance(member, decorators.CapabilityProperty)]
        return functools.reduce(set.union, capability_classes, set())

    @classmethod
    def has_capabilities(cls, capability_names):
        """Check whether this device class supports all of the given capabilities.

        Args:
            capability_names (list): names of capabilities to check for.

        Note:
            capability names are strings. They can be:
            - capability names ("keypad"),
            - capability interface names ("keypadbase"),
            - capability flavor names ("keypaddefault").
            If an interface name or capability name is specified, the behavior is identical:
                any capability flavor which implements the given interface will match.
            If a flavor name is specified, only that capability flavor will match.
            Different kinds of capability names can be used together (["wifi", "keypaddefault"]).

        Returns:
            bool: True if all of the given capabilities are supported by this device class,
                  False otherwise.

        Raises:
            GazooDeviceError: invalid type of capability_names argument OR
                             one of the capability names provided isn't recognized by GDM.
        """
        valid_capability_names_types = (list, tuple, set)
        if not isinstance(capability_names, valid_capability_names_types):
            raise errors.GazooDeviceError(
                "Invalid type of capability_names. Expected one of: {}, found: {}."
                .format(valid_capability_names_types, type(capability_names)))
        if not all(isinstance(cap_name, str) for cap_name in capability_names):
            raise errors.GazooDeviceError(
                "All capability names must be of string type. Found: {}.".format(capability_names))

        capabilities = []  # Interface or flavor classes
        for cap_name in capability_names:
            cap_name = cap_name.lower()
            if cap_name in supported_classes.SUPPORTED_CAPABILITY_FLAVORS:
                interface_or_flavor = supported_classes.SUPPORTED_CAPABILITY_FLAVORS[cap_name]
            elif cap_name in supported_classes.SUPPORTED_CAPABILITY_INTERFACES:
                interface_or_flavor = supported_classes.SUPPORTED_CAPABILITY_INTERFACES[cap_name]
            elif cap_name in supported_classes.SUPPORTED_CAPABILITIES:
                interface_name = supported_classes.SUPPORTED_CAPABILITIES[cap_name]
                interface_or_flavor = supported_classes.SUPPORTED_CAPABILITY_INTERFACES[
                    interface_name]
            else:
                msg = "\n".join([
                    "Capability {} is not recognized.".format(cap_name),
                    "Supported capability interfaces: {}".format(
                        supported_classes.SUPPORTED_CAPABILITY_INTERFACES.keys()),
                    "Supported capability flavors: {}".format(
                        supported_classes.SUPPORTED_CAPABILITY_FLAVORS.keys()),
                    "Supported capabilities: {}".format(
                        supported_classes.SUPPORTED_CAPABILITIES.keys())
                ])
                raise errors.GazooDeviceError(msg)
            capabilities.append(interface_or_flavor)

        supported_capabilities = cls.get_supported_capability_flavors()
        for requested_capability in capabilities:
            if not any(issubclass(supported_capability, requested_capability)
                       for supported_capability in supported_capabilities):
                return False
        return True

    @decorators.LogDecorator(logger, decorators.DEBUG)
    def make_device_ready(self, setting="on"):
        """Check device readiness and attempt recovery if allowed.

        Args:
            setting(str): 'on' | 'off' | 'check_only'.

        Raises:
            GazooDeviceError: if device communication fails.

        Note:
            If setting is 'check_only', will skip recovery.
            If setting is 'off', will skip check_device_ready and recover.
        """
        if setting == "off":
            return

        # check if the device is ready
        try:
            self.check_device_ready()
        except errors.GazooDeviceError as err:
            if setting == "check_only":
                logger.info("{} skipping device recovery", self.name)
                raise

            # attempt to recover the device
            logger.info("{} failed check_device_ready with {}".format(self.name, repr(err)))
            self.recover(err)

            # check if the device is ready after recovery
            logger.info("{} re-checking device readiness after recovery attempt", self.name)
            try:
                self.check_device_ready()
            except errors.GazooDeviceError as err:
                logger.info("{} failed check_device_ready with {}".format(self.name, repr(err)))
                raise

    @decorators.LogDecorator(logger, decorators.DEBUG)
    def reset_all_capabilities(self):
        """Resets all capabilities which have been initialized by deleting them.

        Capabilities will be re-initialized on next use (when they're accessed).
        """
        for capability_class in self.get_supported_capability_flavors():
            self.reset_capability(capability_class)

    @decorators.LogDecorator(logger, decorators.DEBUG)
    def reset_capability(self, capability_class):
        """Resets the capability if it's already initialized by deleting it.

        The capability will be re-initialized on next capability use (when it's accessed).

        Args:
            capability_class (object): class of the capability to reset.
        """
        capability_name = self._get_private_capability_name(capability_class)
        if hasattr(self, capability_name):
            delattr(self, capability_name)

    @decorators.LogDecorator(logger, decorators.DEBUG)
    def set_property(self, prop, value):
        """Set a settable property.

        Args:
            prop (str): property name
            value (object): value of the property name.

        Raises:
            ValueError: if property is persistent or dynamic.
        """
        if prop in self.get_persistent_property_names():
            raise ValueError("{}'s {} is a persistent property and not settable. "
                             "Redetect device if it's wrong.".format(self.name, prop))
        if prop in self.get_dynamic_property_names():
            raise ValueError("{}'s {} is a dynamic property and not settable. ".format(
                self.name, prop))
        self.props['optional'][prop] = value

    def command_with_regex(self,
                           command,
                           regex,
                           command_fn,
                           regex_group=1,
                           raise_error=False,
                           tries=1,
                           **command_fn_kwargs):
        """Sends a command, searches for a regex in the response, and returns a match group.

        Args:
            command (str): command to issue.
            regex (str): regular expression with one or more capturing groups.
            command_fn (func): the function used to send the command.
            regex_group (int): number of regex group to return.
            raise_error (bool): whether or not to raise error if unable to find a match.
            tries (int): how many times to try executing the command before failing.
            **command_fn_kwargs (dict): keyword arguments for the command_fn function.

        Returns:
            str: value of the capturing group with index 'regex_group' in the match.

        Raises:
            GazooDeviceError: if command execution fails OR
                             couldn't find the requested group in any of the responses.
        """
        for _ in range(tries):
            try:
                response = command_fn(command, **command_fn_kwargs)
            except errors.GazooDeviceError:
                continue

            match = re.search(regex, response, re.MULTILINE | re.DOTALL)
            if match:
                max_group = match.lastindex
                if max_group is None:
                    max_group = 0

                if regex_group > max_group:
                    logger.warning(
                        "{}: requested group index ({}) exceeds index of last matched group ({}). "
                        "Matched groups: {}, response: {!r}, regex: {!r}."
                        .format(self.name, regex_group, max_group, match.groups(), response,
                                regex))
                else:
                    return str(match.group(regex_group))

        msg = "{} unable to retrieve {} from {} after {} tries".format(self.name, regex, command,
                                                                       tries)

        if raise_error:
            raise errors.GazooDeviceError(msg)
        else:
            logger.warning(msg)
            return ""

    @decorators.health_check
    def check_ping_responsiveness(self):
        """Check if the auxiliary device responds to pings.

        Raises:
            DeviceNotResponsiveError: if no response to ping before the timeout.
        """
        try:
            common_utils.retry(self._ping, is_successful=bool, timeout=self._PING_TIMEOUT,
                               reraise=True)
        except Exception as err:
            raise errors.DeviceNotResponsiveError(
                self.name, "failed to respond to ping", timeout=self._PING_TIMEOUT,
                details=str(err))

    def _execute_health_check_methods(self, health_checks):
        """Execute health checks on the device.

        Args:
            health_checks (list): list of methods to execute as health checks.

        Raises:
            CheckDeviceReadyError: if a health check fails.

        Note:
            Order of health check methods matter. Health check execution will stop at the first
            failing health check.
        """
        for step, health_check_method in enumerate(health_checks):
            method_name = health_check_method.__name__
            health_check_name = method_name.replace("_", " ").strip().capitalize()

            try:
                health_check_method()
            except errors.CheckDeviceReadyError:
                logger.info("{} health check {}/{} failed: {}.".format(
                    self.name, step + 1, len(health_checks), health_check_name))
                raise

            logger.info("{} health check {}/{} succeeded: {}.".format(
                self.name, step + 1, len(health_checks), health_check_name))

    def lazy_init(self, capability_class, *args, **kwargs):
        """Provides a lazy instantiation mechanism for capabilities.

        The capability instance will not be created until it is accessed for the first time.
        Subsequent accesses will return the same capability instance.
        In other words, no more than one capability instance will be created.
        If a capability is not accessed, the capability instance is not created.

        Args:
            capability_class (class): capability class to instantiate.
            *args (tuple): positional args to the capability's __init__.
            **kwargs (dict): keyword arguments to the capability's __init__.

        Returns:
            CapabilityBase: initialized capability instance.
        """
        capability_name = self._get_private_capability_name(capability_class)
        if not hasattr(self, capability_name):
            capability_inst = capability_class(*args, **kwargs)
            setattr(self, capability_name, capability_inst)
        return getattr(self, capability_name)

    @classmethod
    def _get_private_capability_name(cls, capability_class):
        """Returns the name of the private attribute to use for a given capability class.

        Args:
            capability_class(class): capability class object.

        Returns:
            str: name of the private capability attribute to be used.
        """
        return "_{}".format(capability_class.get_capability_name())

    @classmethod
    def _get_property_names(cls, property_type):
        """Returns the property names for all public properties of the type."""
        classes = cls.get_supported_capability_flavors().copy()
        classes.add(cls)
        property_names = []
        for a_class in classes:
            prefix = "" if a_class == cls else a_class.get_capability_name() + "."
            for name, member in inspect.getmembers(a_class):
                if isinstance(member, property_type) and not name.startswith("_"):
                    property_names.append("{}{}".format(prefix, name))
        return list(set(property_names))

    def _get_properties(self, property_names):
        """Returns a dictionary of prop, value for each property."""
        property_dict = {}
        for name in property_names:
            value = self.get_property(name)
            if isinstance(value, str) and "does not have a known property" in value:
                continue  # property not supported in current flavor
            property_dict[name] = value

        return property_dict

    def _list_properties_dynamic_auxiliary_device(self):
        """Returns a set of dynamic properties.

        Returns:
            set: dynamic properties that can be queried during get-prop CLI/API operation.
        """
        return set(["connected"])

    def _log_object_lifecycle_event(self, method_name):
        """Logs a message about a lifecycle event of a python object.

        Args:
            method_name (str): name of the method called on the object.
        """
        logger.debug("{} called on {} (id = {}) in process {}.".format(
            method_name, self, id(self), os.getpid()))



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines a process transport by wrapping the subprocess.Popen() constructor."""
from __future__ import absolute_import
import select
import subprocess
import time
import six
from gazoo_device.switchboard import transport_properties
from gazoo_device.switchboard import transport_base


class ProcessTransport(transport_base.TransportBase):
    """Perform communication using a subprocess with the command and args specified."""

    def __init__(self,
                 command,
                 args="",
                 auto_reopen=False,
                 open_on_start=True,
                 working_directory=None):
        """Initialize the ProcessTransport object with the given process properties.

        Args:
            command (str): command to execute
            args (str): list of additional args to pass to command
            auto_reopen (bool): flag indicating transport should be reopened if
                                unexpectedly closed.
            open_on_start (bool): flag indicating transport should be open on
                                  TransportProcess start.
            working_directory (string): Current working directory.
        """

        super(ProcessTransport, self).__init__(auto_reopen, open_on_start)
        self._args = [command] + args.split()

        self._properties.update({
            transport_properties.CLOSE_FDS: True,
        })
        self._process = None
        self._cwd = working_directory

    def is_open(self):
        """Returns True if transport is currently open.

        Returns:
            bool: True if transport is currently open, False otherwise.

        Note:
            If the process unexpected exits this method will return False.
        """

        return (hasattr(self, "_process") and
                self._process is not None and
                self._process.poll() is None)

    def _open(self):
        """Opens or reopens the process using the current property values."""
        if self._is_ready_to_open():
            self._process = subprocess.Popen(
                self._args,
                bufsize=0,
                close_fds=self._properties[transport_properties.CLOSE_FDS],
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                cwd=self._cwd
            )

    def _close(self):
        """The close() implementation checks if process is open differently than is_open().

        Hence this method is unused, but has to be overridden.
        """
        pass

    def close(self):
        """Closes the process."""
        if hasattr(self, "_process") and self._process:
            self._process.stdin.close()
            self._process.stdout.close()
            if self._process.poll() is None:
                self._process.terminate()
                while self._process.poll() is None:
                    time.sleep(0.001)
            self._process = None

    def _read(self, size=1, timeout=None):
        """Returns bytes read up to max_bytes within timeout in seconds specified.

        Args:
            size (int): maximum number of bytes to read within timeout seconds
            timeout (float): maximum seconds to wait to read bytes or indefinitely
                             if timeout=None. Default: None

        Returns:
            str: bytes read from transport or None if no bytes were read
        """
        if timeout:
            return self._read_non_blocking(size, timeout)
        return self._read_data(size)

    def _write(self, data, timeout=None):
        """Writes data provided within timeout in seconds specified.

        Args:
            data (str): bytes to be written out within timeout seconds
            timeout (float): maximum seconds to wait to write bytes or indefinitely
                             if timeout=None. Default: None

        Returns:
            int: number of bytes written or None if no bytes were written
        """
        if isinstance(data, six.text_type):
            data = data.encode("utf-8", errors="replace")

        if timeout:
            return self._write_non_blocking(data, timeout)
        return self._write_data(data)

    def _is_ready_to_open(self):
        return True

    def _read_data(self, size):
        return self._process.stdout.read(size)

    def _read_non_blocking(self, size, timeout):
        count = 0
        result = b""
        end_time = time.time() + timeout
        while count < size and time.time() < end_time:
            time_left = end_time - time.time()
            readable, _, _ = select.select([self._process.stdout],
                                           [],
                                           [],
                                           0 if time_left < 0 else time_left)
            for stream in readable:
                if stream is self._process.stdout:
                    count += 1
                    result += self._read_data(1)
        return result

    def _write_non_blocking(self, data, timeout):
        count = 0
        size = len(data)
        end_time = time.time() + timeout
        while count < size and time.time() < end_time:
            time_left = end_time - time.time()
            _, writable, _ = select.select([],
                                           [self._process.stdin],
                                           [],
                                           0 if time_left < 0 else time_left)
            for stream in writable:
                if stream is self._process.stdin:
                    count += self._write_data(data[count])
        return count

    def _write_data(self, data):
        if isinstance(data, six.integer_types):
            data = six.int2byte(data)
        self._process.stdin.write(data)
        self._process.stdin.flush()
        return len(data)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""ExpectResponse object carries results from calls to GDM expect api.

GDM expect methods, found in switchboard.py, load results into ExpectResponse
to permit calling routines a way to evaluate device execution and behavior.
"""


class ExpectResponse(object):
    """Stores the values of expect Attributes.

    Attributes:
        index (int): index of matching pattern in pattern list (None if timeout).
        before (str): all the characters looked at before the match.
        after (str): all the characters after the first matching character.
        match (re.Match): pattern match object.
        timedout (bool): True if the expect timed out.
        time_elapsed (float): number of seconds between start and finish.
        remaining (list): patterns that have not been matched (after a timeout).
        match_list (list): of re.search Match objects.
    """

    def __init__(self,
                 index,
                 before,
                 after,
                 match,
                 time_elapsed,
                 timedout=False,
                 remaining=None,
                 match_list=None):
        self.index = index
        self.before = before
        self.after = after
        self.match = match
        self.timedout = timedout
        self.time_elapsed = time_elapsed
        self.remaining = [] if remaining is None else remaining
        self.match_list = [] if match_list is None else match_list



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common reusable utility functions."""
import multiprocessing
import os
import time
import weakref
from gazoo_device import errors

# Lists of callable functions with no arguments
_run_before_fork_functions = []
_run_after_fork_in_parent_functions = []

# _Sentinel objects used by multiprocessing.util.register_after_fork. This list keeps the objects
# alive by keeping a strong reference to them. multiprocessing.util stores these in a
# weakref.WeakValueDictionary, so if a strong reference is not stored somewhere, the sentinel
# will be destroyed, removing the function registered with it from the list of "after fork"
# functions (b/154659535).
_register_after_fork_sentinels = []


class MethodWeakRef(object):
    """Allows creating weak references to instance methods.

    Note:
        using weakref.ref() on an instance method directly returns a dead reference.
        See https://stackoverflow.com/questions/599430.
    """

    def __init__(self, instance_method):
        """Create a weak reference to an instance (bound) method.

        Args:
            instance_method (method): instance (bound) method.
        """
        self._func = instance_method.__func__
        self._instance_weakref = weakref.ref(instance_method.__self__)

    def __call__(self, *args, **kwargs):
        """Calls the instance method if the instance is still alive."""
        instance = self._instance_weakref()
        if instance is not None:
            self._func(instance, *args, **kwargs)


class _Sentinel(object):
    """Empty object which supports weakrefs."""
    pass


def _default_is_successful(_):
    return True


def not_func(val):
    """Takes an input value and returns bool value after performing a negate operation.

    Args:
        val (object): any python object

    Returns:
        bool: True if bool(val) is False and False if bool(val) is True
    """
    return not bool(val)


def retry(func, func_args=(), func_kwargs=None, is_successful=_default_is_successful,
          timeout=10, interval=1, reraise=True, exc_type=errors.CommunicationTimeoutError):
    """Wait until either execution of func() succeeds or timeout is reached.

    Success of execution of func() is determined by is_successful() function, which should return
        True on successful execution of func().

    Args:
        func (function): function to execute
        func_args (tuple): positional arguments to the function
        func_kwargs (dict): keyword arguments to the function
        is_successful (function): function which takes in the result of func() and returns
                                  whether function execution should be considered successful.
                                  To indicate success, return True.
                                  Defaults to always returning True.
        timeout (int): if no run of func() succeeds in this time period, raise an error.
        interval (float): how long to wait between retries of func().
        reraise (bool): whether to re-raise exceptions in func() or not.
                        If True, will re-raise any exceptions from func().
                        If False, considers execution of func() a failure if an Exception
                        is raised. is_successful() will NOT be called if an Exception occurs.
        exc_type (class): type of exception to raise when timeout is reached.
                          Note that the class constructor will be called with just 1 argument.

    Returns:
        object: return value of first successful func() call.

    Raises:
        Exception: if timeout is reached. The default exception type is errors.TimeoutError. OR
                   if an Exception occurs in func() and reraise is False.
    """
    if func_kwargs is None:
        func_kwargs = {}

    tried_times = 0
    start_time = time.time()
    end_time = start_time + timeout

    while time.time() < end_time:
        exception_occurred = False
        tried_times += 1
        try:
            func_result = func(*func_args, **func_kwargs)
        except Exception:
            if reraise:
                raise
            else:
                exception_occurred = True

        if not exception_occurred and is_successful(func_result):
            return func_result

        time.sleep(interval)

    time_elapsed = time.time() - start_time
    raise exc_type("Timeout in {}s. Tried calling {} {} times."
                   .format(time_elapsed, func.__name__, tried_times))


def generate_name(_object):
    """Generates a snake_case name (str) for the given object.

    Args:
        _object (object): any python object which has a non-empty __name__ attribute.

    Raises:
        ValueError: _object does not have __name__ attribute, or __name__ is "".

    Returns:
        str: snake_case name for the object.
    """
    object_name = getattr(_object, "__name__", "")
    if not object_name:
        raise ValueError("Object {} must have a non-empty __name__ attribute.".format(_object))

    if "_" in object_name:  # Presumably already snake_case
        return object_name.lower()
    else:  # Presumably TitleCase
        return title_to_snake_case(object_name)


def get_value_from_json(json_data, key_sequence, raise_if_absent=True):
    """Extracts a value from a JSON dictionary via provided series of keys.

         This function allows easy extraction of a value stored in a nested
         dictionary structure, with validation of the existence of the necessary
         keys at each level of nesting.

    Args:
        json_data (dict): the response object from which the value will be extracted.
        key_sequence (list): an ordered list of string keys representing the path through
                             the dictionary necessary to reach the desired value.
                             Ex. to extract foo['baz']['a'] the call would look like:
                             get_value_from_json(foo, ['baz', 'a'])  and would return
                             'some text' given the example data above.
        raise_if_absent (bool): if False, return None if a key is not present.
                                if True, raise KeyError if a key is not present.

    Returns:
        object: the requested JSON field.

    Raises:
        KeyError: If the provided response message doesn't contain the necessary keys/values
                  AND if raise_if_absent is True.
    """
    current_dict = json_data
    key_text = ''
    for key in key_sequence:
        if key in current_dict:
            current_dict = current_dict[key]
        else:
            if raise_if_absent:
                raise KeyError(
                    "Unable to find key '{}' in json_data{}: json_data = {}"
                    .format(key, key_text, json_data))
            else:
                return None
        key_text += "['{}']".format(key)
    return current_dict


def register_at_fork(before=None, after_in_parent=None, after_in_child=None):
    """Workaround to simulate os.register_at_fork() behavior in all Python versions.

    Args:
        before (func): function to call prior to os.fork() call.
        after_in_parent (func): function to call after os.fork() returns in the parent process.
        after_in_child (func): function to call after os.fork() returns in the child process.

    Note:
        For Python >= 3.7, this delegates to os.register_at_fork().
        For Python < 3.7, the functions need to be run manually via
            common_utils.run_before_fork() prior to os.fork() and
            common_utils.run_after_fork_in_parent() after os.fork() returns in the parent process.
        Note that os.fork() is used internally by the "multiprocessing" library:
            multiprocessing.Process().start() and multiprocessing.Manager() both use os.fork().
    """
    if hasattr(os, "register_at_fork"):
        os.register_at_fork(before=before,
                            after_in_parent=after_in_parent,
                            after_in_child=after_in_child)
    else:
        global _run_before_fork_functions
        global _run_after_fork_in_parent_functions
        global _register_after_fork_sentinels

        if callable(before):
            _run_before_fork_functions.append(before)
        if callable(after_in_parent):
            _run_after_fork_in_parent_functions.append(after_in_parent)
        if callable(after_in_child):

            def _consume_sentinel_wrapper(_):
                """Consumes the sentinel argument passed into the function."""
                after_in_child()

            sentinel = _Sentinel()
            _register_after_fork_sentinels.append(sentinel)
            multiprocessing.util.register_after_fork(sentinel, _consume_sentinel_wrapper)


def run_after_fork_in_parent():
    """Run functions registered to run in parent after os.fork() calls."""
    for func in _run_after_fork_in_parent_functions:
        func()


def run_before_fork():
    """Run functions registered to run before os.fork() calls."""
    for func in _run_before_fork_functions:
        func()


def title_to_snake_case(s):
    """Convert TitleCase string to snake_case.

    Args:
        s (str): TitleCase string.

    Returns:
        str: snake_case string.

    Raises:
        ValueError: provided string contains underscores.

    Note:
        consecutive capital characters are supported ("ABc"), but underscores are not.
    """
    if "_" in s:
        raise ValueError("{} is not a TitleCase string (found underscores).".format(s))

    word_starts = [pos for pos in range(len(s)) if _is_new_word(s, pos)] + [len(s)]
    words = [s[word_starts[idx]:word_starts[idx + 1]].lower()
             for idx in range(len(word_starts) - 1)]
    return "_".join(words)


def _is_new_word(s, pos):
    """Returns whether a new words starts at s[pos] in a TitleCase string."""
    return (pos == 0 or
            (s[pos].isupper() and (
                not s[pos - 1].isupper()
                or (pos + 1 < len(s) and not s[pos + 1].isupper()))))
